#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Sequence
from logging import Logger
from typing import Any, cast, TYPE_CHECKING, Union

# TODO[@mgarrard, @drfreund]: Remove this when we streamline `apply_input
# constructors`, such that they no longer need to access individual
# input constructor purposes.
import ax.generation_strategy as gs_module  # @manual

from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.observation import ObservationFeatures
from ax.core.trial_status import TrialStatus
from ax.exceptions.core import AxError, UserInputError
from ax.exceptions.generation_strategy import GenerationStrategyRepeatedPoints
from ax.exceptions.model import ModelError
from ax.generation_strategy.best_model_selector import BestModelSelector

if TYPE_CHECKING:
    from ax.generation_strategy.generation_node_input_constructors import (
        InputConstructorPurpose,
        NodeInputConstructors,
    )
    from ax.generation_strategy.generation_strategy import GenerationStrategy

from ax.adapter.base import Adapter
from ax.adapter.registry import (
    _extract_model_state_after_gen,
    GeneratorRegistryBase,
    Generators,
)
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.generation_strategy.transition_criterion import (
    AutoTransitionAfterGen,
    MaxGenerationParallelism,
    MinTrials,
    TransitionCriterion,
    TrialBasedCriterion,
)
from ax.utils.common.base import SortableBase
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from ax.utils.common.serialization import SerializationMixin
from pyre_extensions import none_throws


logger: Logger = get_logger(__name__)

TModelFactory = Callable[..., Adapter]
MISSING_MODEL_SELECTOR_MESSAGE = (
    "A `BestModelSelector` must be provided when using multiple "
    "`GeneratorSpec`s in a `GenerationNode`. After fitting all `GeneratorSpec`s, "
    "the `BestModelSelector` will be used to select the `GeneratorSpec` to "
    "use for candidate generation."
)
MAX_GEN_ATTEMPTS = 5
MAX_GEN_ATTEMPTS_EXCEEDED_MESSAGE = (
    f"GenerationStrategy exceeded `MAX_GEN_ATTEMPTS` of {MAX_GEN_ATTEMPTS} while "
    "trying to generate a unique parameterization. This indicates that the search "
    "space has likely been fully explored, or that the sweep has converged."
)
DEFAULT_FALLBACK: dict[type[Exception], GeneratorSpec] = {
    cast(type[Exception], GenerationStrategyRepeatedPoints): GeneratorSpec(
        generator_enum=Generators.SOBOL, model_key_override="Fallback_Sobol"
    )
}


class GenerationNode(SerializationMixin, SortableBase):
    """Base class for GenerationNode, capable of fitting one or more model specs under
    the hood and generating candidates from them.

    Args:
        node_name: A unique name for the GenerationNode. Used for storage purposes.
        generator_specs: A list of GeneratorSpecs to be selected from for generation
            in this GenerationNode.
        best_model_selector: A ``BestModelSelector`` used to select the
            ``GeneratorSpec`` to generate from in ``GenerationNode`` with
            multiple ``GeneratorSpec``s.
        should_deduplicate: Whether to deduplicate the parameters of proposed arms
            against those of previous arms via rejection sampling. If this is True,
            the GenerationStrategy will discard generator runs produced from the
            GenerationNode that has `should_deduplicate=True` if they contain arms
            already present on the experiment and replace them with new generator runs.
            If no generator run with entirely unique arms could be produced in 5
            attempts, a `GenerationStrategyRepeatedPoints` error will be raised, as we
            assume that the optimization converged when the model can no longer suggest
            unique arms.
        transition_criteria: List of TransitionCriterion, each of which describes a
            condition that must be met before completing a GenerationNode. All `is_met`
            must evaluateTrue for the GenerationStrategy to move on to the next
            GenerationNode.
        input_constructors: A dictionary mapping input constructor purpose enum to the
            input constructor enum. Each input constructor maps to a method which
            encodes the logic for determining dynamic inputs to the ``GenerationNode``
        trial_type: Specifies the type of trial to generate, is limited to either
            ``Keys.SHORT_RUN`` or ``Keys.LONG_RUN`` for now. If not specified, will
            default to None and not be used during generation.
        previous_node_name: The previous ``GenerationNode`` name in the
            ``GenerationStrategy``, if any. Initialized to None for all nodes, and is
            set during transition from one ``GenerationNode`` to the next. Can be
            overwritten if multiple transitions occur between nodes, and will always
            store the most recent previous ``GenerationNode`` name.
        should_skip: Whether to skip this node during generation time. Defaults to
            False, and can only currently be set to True via ``NodeInputConstructors``
        fallback_specs: Optional dict mapping expected exception types to `ModelSpec`
            fallbacks used when gen fails.

    Note for developers: by "model" here we really mean an Ax Adapter object, which
    contains an Ax Model under the hood. We call it "model" here to simplify and focus
    on explaining the logic of GenerationStep and GenerationStrategy.
    """

    # Required options:
    generator_specs: list[GeneratorSpec]
    # TODO: Move `should_deduplicate` to `GeneratorSpec` if possible, and make optional
    should_deduplicate: bool
    _node_name: str

    # Optional specifications
    _generator_spec_to_gen_from: GeneratorSpec | None = None
    # TODO: @mgarrard should this be a dict criterion_class name -> criterion mapping?
    _transition_criteria: Sequence[TransitionCriterion]
    _input_constructors: dict[
        InputConstructorPurpose,
        NodeInputConstructors,
    ]
    _previous_node_name: str | None = None
    _trial_type: str | None = None
    _should_skip: bool = False
    fallback_specs: dict[type[Exception], GeneratorSpec]

    # [TODO] Handle experiment passing more eloquently by enforcing experiment
    # attribute is set in generation strategies class
    _generation_strategy: None | GenerationStrategy = None

    def __init__(
        self,
        node_name: str,
        generator_specs: list[GeneratorSpec],
        best_model_selector: BestModelSelector | None = None,
        should_deduplicate: bool = False,
        transition_criteria: Sequence[TransitionCriterion] | None = None,
        input_constructors: None
        | (
            dict[
                InputConstructorPurpose,
                NodeInputConstructors,
            ]
        ) = None,
        previous_node_name: str | None = None,
        trial_type: str | None = None,
        should_skip: bool = False,
        fallback_specs: dict[type[Exception], GeneratorSpec] | None = None,
    ) -> None:
        self._node_name = node_name
        # Check that the model specs have unique model keys.
        model_keys = {generator_spec.model_key for generator_spec in generator_specs}
        if len(model_keys) != len(generator_specs):
            raise UserInputError(
                "Model keys must be unique across all model specs in a GenerationNode."
            )
        if len(generator_specs) > 1 and best_model_selector is None:
            raise UserInputError(MISSING_MODEL_SELECTOR_MESSAGE)
        if trial_type is not None and (
            trial_type != Keys.SHORT_RUN and trial_type != Keys.LONG_RUN
        ):
            raise NotImplementedError(
                f"Trial type must be either {Keys.SHORT_RUN} or {Keys.LONG_RUN},"
                f" got {trial_type}."
            )
        # If possible, assign `_generator_spec_to_gen_from` right away, for use in
        # `__repr__`
        if len(generator_specs) == 1:
            self._generator_spec_to_gen_from = generator_specs[0]
        self.generator_specs = generator_specs
        self.best_model_selector = best_model_selector
        self.should_deduplicate = should_deduplicate
        self._transition_criteria = (
            transition_criteria if transition_criteria is not None else []
        )
        self._input_constructors = (
            input_constructors if input_constructors is not None else {}
        )
        self._previous_node_name = previous_node_name
        self._trial_type = trial_type
        self._should_skip = should_skip
        self.fallback_specs = (
            fallback_specs if fallback_specs is not None else DEFAULT_FALLBACK
        )

    def __eq__(self, other: SortableBase) -> bool:
        return SortableBase.__eq__(self, other)

    @property
    def node_name(self) -> str:
        """Returns the unique name of this GenerationNode"""
        return self._node_name

    @property
    def generator_spec_to_gen_from(self) -> GeneratorSpec:
        """Returns the cached `_generator_spec_to_gen_from` or gets it from
        `_pick_fitted_adapter_to_gen_from` and then caches and returns it
        """
        if self._generator_spec_to_gen_from is not None:
            return self._generator_spec_to_gen_from

        self._generator_spec_to_gen_from = self._pick_fitted_adapter_to_gen_from()
        return self._generator_spec_to_gen_from

    @property
    def model_to_gen_from_name(self) -> str | None:
        """Returns the name of the model that will be used for gen, if there is one.
        Otherwise, returns None.
        """
        if self._generator_spec_to_gen_from is not None:
            return self._generator_spec_to_gen_from.model_key
        else:
            return None

    @property
    def generation_strategy(self) -> GenerationStrategy:
        """Returns a backpointer to the GenerationStrategy, useful for obtaining the
        experiment associated with this GenerationStrategy"""
        # TODO: @mgarrard remove this property once we make experiment a required
        # argument on GenerationStrategy
        if self._generation_strategy is None:
            raise ValueError(
                "Generation strategy has not been initialized on this node."
            )
        return none_throws(self._generation_strategy)

    @property
    def transition_criteria(self) -> Sequence[TransitionCriterion]:
        """Returns the sequence of TransitionCriteria that will be used to determine
        if this GenerationNode is complete and should transition to the next node.
        """
        return [] if self._transition_criteria is None else self._transition_criteria

    @property
    def input_constructors(
        self,
    ) -> dict[
        InputConstructorPurpose,
        NodeInputConstructors,
    ]:
        """Returns the input constructors that will be used to determine any dynamic
        inputs to this ``GenerationNode``.
        """
        return self._input_constructors if self._input_constructors is not None else {}

    @property
    def experiment(self) -> Experiment:
        """Returns the experiment associated with this GenerationStrategy"""
        return self.generation_strategy.experiment

    @property
    def is_completed(self) -> bool:
        """Returns True if this GenerationNode is complete and should transition to
        the next node.
        """
        # TODO: @mgarrard this logic more robust and general
        # We won't mark a node completed if it has an AutoTransitionAfterGen criterion
        # as this is typically used in cyclic generation strategies
        return self.should_transition_to_next_node(raise_data_required_error=False)[
            0
        ] and not any(
            isinstance(tc, AutoTransitionAfterGen) for tc in self.transition_criteria
        )

    @property
    def previous_node(self) -> GenerationNode | None:
        """Returns the previous ``GenerationNode``, if any."""
        return (
            self.generation_strategy.nodes_dict[self._previous_node_name]
            if self._previous_node_name is not None
            else None
        )

    @property
    def _unique_id(self) -> str:
        """Returns a unique (w.r.t. parent class: ``GenerationStrategy``) id
        for this GenerationNode. Used for SQL storage.
        """
        return self.node_name

    @property
    def _fitted_adapter(self) -> Adapter | None:
        """Private property to return optional fitted_adapter from
        self.generator_spec_to_gen_from for convenience. If no model is fit,
        this will return None.
        """
        try:
            # Using the private attribute since using the non-private `fitted_adapter`
            # property will raise a UserInputError if there is no fitted model.
            return self.generator_spec_to_gen_from._fitted_adapter
        except ModelError:
            # ModelError is raised if there are no fitted adapters to select from.
            return None

    def __repr__(self) -> str:
        """String representation of this ``GenerationNode`` (note that it
        will abridge some aspects of ``TransitionCriterion`` and
        ``GeneratorSpec`` attributes).
        """
        str_rep = f"{self.__class__.__name__}"
        str_rep += f"(node_name='{self.node_name}'"
        str_rep += ", generator_specs="
        generator_spec_str = (
            ", ".join([spec._brief_repr() for spec in self.generator_specs])
            .replace("\n", " ")
            .replace("\t", "")
        )
        str_rep += f"[{generator_spec_str}]"
        str_rep += (
            f", transition_criteria={str(self._brief_transition_criteria_repr())}"
        )
        return f"{str_rep})"

    def _fit(
        self,
        experiment: Experiment,
        data: Data | None = None,
        **kwargs: Any,
    ) -> None:
        """Fits the specified models to the given experiment + data using
        the model kwargs set on each corresponding model spec and the kwargs
        passed to this method.

        NOTE: During fitting of the ``GeneratorSpec``, state of this ``GeneratorSpec``
        after its last candidate generation is extracted from the last
        ``GeneratorRun`` it produced (if any was captured in
        ``GeneratorRun.model_state_after_gen``) and passed into ``GeneratorSpec.fit``
        as keyword arguments.

        Args:
            experiment: The experiment to fit the model to.
            data: The experiment data used to fit the model, optional (if not specified
                will use ``experiment.lookup_data()``, extracted in ``Adapter``).
            kwargs: Additional keyword arguments to pass to the model's
                ``fit`` method. NOTE: Local kwargs take precedence over the ones
                stored in ``GeneratorSpec.model_kwargs``.
        """
        self._generator_spec_to_gen_from = None
        for generator_spec in self.generator_specs:
            try:
                # Stores the fitted model as `generator_spec._fitted_adapter`.
                generator_spec.fit(
                    experiment=experiment,
                    data=data,
                    **{
                        **self._get_model_state_from_last_generator_run(
                            generator_spec=generator_spec
                        ),
                        **kwargs,
                    },
                )
            except Exception as e:
                if len(self.generator_specs) == 1:
                    # If no other model specs, raise.
                    raise
                # If there are other model specs, try to handle gracefully.
                logger.exception(
                    "Model fitting failed for `GeneratorSpec` "
                    f"{generator_spec.model_key}. Original error message: {e}."
                )
                # Discard any previously fitted models for this spec.
                generator_spec._fitted_adapter = None

    def gen(
        self,
        *,
        experiment: Experiment,
        pending_observations: dict[str, list[ObservationFeatures]] | None,
        skip_fit: bool = False,
        data: Data | None = None,
        **gs_gen_kwargs: Any,
    ) -> GeneratorRun | None:
        """This method generates candidates using `self._gen` and handles deduplication
        of generated candidates if `self.should_deduplicate=True`.

        NOTE: This method returning ``None`` indicates that this node should be skipped
            by its generation strategy.
        NOTE: ``n`` argument will be passed through this node's input constructors,
            which might modify the number of arms this node will produce. Also, some
            underlying generators  may ignore the ``n`` argument and produce a
            model-determined number of arms. In that case this method will also output
            a generator run with number of arms that may differ from ``n``.

        Args:
            experiment: The experiment to generate candidates for.
            data: Optional override for the experiment data used to generate candidates;
                if not specified, will use ``experiment.lookup_data()`` (extracted in
                ``Adapter``).
            pending_observations: A map from metric name to pending
                observations for that metric, used by some models to avoid
                resuggesting points that are currently being evaluated.
            model_gen_kwargs: Keyword arguments, passed through to
                ``ModelSpec.gen``; these override any pre-specified in
                ``ModelSpec.model_gen_kwargs``. Often will contain ``n``.

        Returns:
            A ``GeneratorRun`` containing the newly generated candidates or ``None``
            if this node is not in a correct state to generate candidates and should
            be skipped (e.g. if its input constructor for the ``n`` argument specifies
            that it should generate 0 candidate arms given the current experiment
            state and user inputs).
        """
        # TODO: Consider returning "should skip" from apply input constructors
        input_constructor_values = self.apply_input_constructors(
            experiment=experiment,
            gen_kwargs=gs_gen_kwargs,
        )
        # If during input constructor application we determined that we should skip
        # this node, return early.
        if self._should_skip:
            logger.debug(f"Skipping generation for node {self.node_name}.")
            return None

        if not skip_fit:
            self._fit(experiment=experiment, data=data)
        model_gen_kwargs = gs_gen_kwargs.copy()
        model_gen_kwargs.update(input_constructor_values)
        try:
            # Generate from the main generator on this node. If deduplicating,
            # keep generating until each of `generator_run.arms` is not a
            # duplicate of a previous active arm (e.g. not from a failed trial)
            # on the experiment.
            gr = self._gen_maybe_deduplicate(
                experiment=experiment,
                data=data,
                pending_observations=pending_observations,
                **model_gen_kwargs,
            )
        except Exception as e:
            gr = self._try_gen_with_fallback(
                exception=e,
                experiment=experiment,
                data=data,
                pending_observations=pending_observations,
                **model_gen_kwargs,
            )

        gr._generation_node_name = self.node_name
        # TODO: @mgarrard determine a more refined way to indicate trial type
        if self._trial_type is not None:
            gen_metadata = gr.gen_metadata if gr.gen_metadata is not None else {}
            gen_metadata["trial_type"] = self._trial_type
            gr._gen_metadata = gen_metadata
        return gr

    def _gen(
        self,
        experiment: Experiment,
        n: int | None,
        pending_observations: dict[str, list[ObservationFeatures]] | None,
        data: Data | None,
        **model_gen_kwargs: Any,
    ) -> GeneratorRun:
        """Picks a fitted model, from which to generate candidates (via
        ``self._pick_fitted_adapter_to_gen_from``) and generates candidates
        from it. Uses the ``model_gen_kwargs`` set on the selected ``GeneratorSpec``
        alongside any kwargs passed in to this function (with local kwargs)
        taking precedent.

        Args:
            experiment: The experiment to generate candidates for.
            data: The experiment data used to generate candidates (usually determined
                via ``experiment.lookup_data()`` in ``GenerationStrategy.gen``, but
                could be manually specified by a user).
            n: Optional integer representing how many arms should be in the generator
                run produced by this method. When this is ``None``, ``n`` will be
                determined by the ``GeneratorSpec`` that we are generating from.
            pending_observations: A map from metric name to pending
                observations for that metric, used by some models to avoid
                resuggesting points that are currently being evaluated.
            model_gen_kwargs: Keyword arguments, passed through to
                ``GeneratorSpec.gen``;these override any pre-specified in
                ``GeneratorSpec.model_gen_kwargs``.

        Returns:
            A ``GeneratorRun`` containing the newly generated candidates.
        """
        generator_spec = self.generator_spec_to_gen_from
        if n is None and generator_spec.model_gen_kwargs:
            # If `n` is not specified, ensure that the `None` value does not
            # override the one set in `generator_spec.model_gen_kwargs`.
            n = generator_spec.model_gen_kwargs.get("n", None)
        return generator_spec.gen(
            experiment=experiment,
            data=data,
            n=n,
            # For `pending_observations`, prefer the input to this function, as
            # `pending_observations` are dynamic throughout the experiment and thus
            # unlikely to be specified in `generator_spec.model_gen_kwargs`.
            pending_observations=pending_observations,
            **model_gen_kwargs,
        )

    def _gen_maybe_deduplicate(
        self,
        experiment: Experiment,
        n: int | None,
        pending_observations: dict[str, list[ObservationFeatures]] | None,
        data: Data | None,
        **model_gen_kwargs: Any,
    ) -> GeneratorRun:
        """Attempts to generate candidates from the main ``GeneratorSpec``
        on this ``GenerationNode``, with deduplication if
        ``self.should_deduplicate=True``. If maximum number of deduplication
        attempts is exceeded, raises ``GenerationStrategyRepeatedPoints``.

        NOTE: Should only ever be called from ``GenerationNode.gen``.
        """
        n_gen_draws = 0
        dedup_against_arms = experiment.arms_by_signature_for_deduplication
        # Keep generating until each of `generator_run.arms` is not a duplicate
        # of a previous arm, if `should_deduplicate is True`
        while n_gen_draws < MAX_GEN_ATTEMPTS:
            n_gen_draws += 1
            gr = self._gen(
                experiment=experiment,
                data=data,
                n=n,
                pending_observations=pending_observations,
                **model_gen_kwargs,
            )
            if not self.should_deduplicate or not dedup_against_arms:
                return gr  # Not deduplicationg.
            if all(arm.signature not in dedup_against_arms for arm in gr.arms):
                return gr  # Generated a set of all-non-duplicate arms.
            logger.debug(
                "The generator run produced duplicate arms. Re-running the "
                "generation step in an attempt to deduplicate. Candidates "
                f"produced in the last generator run: {gr.arms}."
            )

        raise GenerationStrategyRepeatedPoints(MAX_GEN_ATTEMPTS_EXCEEDED_MESSAGE)

    def _try_gen_with_fallback(
        self,
        exception: Exception,
        experiment: Experiment,
        n: int | None,
        data: Data | None,
        pending_observations: dict[str, list[ObservationFeatures]] | None,
        **model_gen_kwargs: Any,
    ) -> GeneratorRun:
        """Attempts to generate candidates from the fallback ``GeneratorSpec``
        on this ``GenerationNode``. Identifies the correct fallback based on the
        type of ``Exception`` thrown by ``_gen_maybe_deduplicate``.

        NOTE: Should only ever be called from ``GenerationNode.gen``.
        """
        error_type = type(exception)
        if error_type not in self.fallback_specs:
            raise exception

        # identify fallback model to use
        fallback_model = self.fallback_specs[error_type]
        logger.warning(
            f"gen failed with error {exception}, "
            "switching to fallback model with generator_enum "
            f"{fallback_model.generator_enum}"
        )

        # Fit fallback model using information from the experiment as ground truth.
        fallback_model.fit(
            experiment=experiment,
            **self._get_model_state_from_last_generator_run(
                generator_spec=fallback_model
            ),
        )
        # Switch _generator_spec_to_gen_from to a fallback spec
        self._generator_spec_to_gen_from = fallback_model
        gr = self._gen(
            experiment=experiment,
            data=data,
            n=n,
            pending_observations=pending_observations,
            **model_gen_kwargs,
        )
        return gr

    def _get_model_state_from_last_generator_run(
        self, generator_spec: GeneratorSpec
    ) -> dict[str, Any]:
        """Get the fit args from the last generator run for the model being fit.

        Args:
            generator_spec: The model spec to get the fit args for.

        Returns:
            A dictionary of fit args extracted from the last generator run
            that was generated by the model being fit.
        """
        if self._generation_strategy is None:
            # If there is no GS, we cannot access the previous GRs.
            return {}
        curr_model = generator_spec.generator_enum
        # Find the last GR that was generated by the model being fit.
        grs = self.generation_strategy._generator_runs
        for gr in reversed(grs):
            if (
                gr._generation_node_name == self.node_name
                and gr._model_key == generator_spec.model_key
            ):
                # Extract the fit args from the GR.
                return _extract_model_state_after_gen(
                    generator_run=gr,
                    model_class=curr_model.model_class,
                )
        # No previous GR from this model.
        return {}

    # ------------------------- Model selection logic helpers. -------------------------

    def _pick_fitted_adapter_to_gen_from(self) -> GeneratorSpec:
        """Select one model to generate from among the fitted models on this
        generation node.

        NOTE: In base ``GenerationNode`` class, this method does the following:
          1. if this ``GenerationNode`` has an associated ``BestModelSelector``,
             use it to select one model to generate from among the fitted models
             on this generation node.
          2. otherwise, ensure that this ``GenerationNode`` only contains one
             `GeneratorSpec` and select it.
        """
        if self.best_model_selector is None:
            if (
                len(self.generator_specs) != 1
            ):  # pragma: no cover -- raised in __init__.
                raise UserInputError(MISSING_MODEL_SELECTOR_MESSAGE)
            return self.generator_specs[0]

        fitted_specs = [
            # Only select between models that were successfully fit.
            spec
            for spec in self.generator_specs
            if spec._fitted_adapter is not None
        ]
        if len(fitted_specs) == 0:
            raise ModelError(
                "No fitted models were found on the `GeneratorSpec`s. "
                "This can be caused by model fitting errors, which should be "
                "diagnosed by following the exception logs produced earlier."
            )
        try:
            best_model = none_throws(self.best_model_selector).best_model(
                generator_specs=fitted_specs,
            )
            return best_model
        except Exception as e:
            logger.warning(
                "The `BestModelSelector` raised an error when selecting the best "
                "generator. This can happen if the generator ran into issues during "
                "computing the relevant diagnostics, such as insufficient training "
                "data. Returning the first generator that was successfully fit. "
                f"Original error message: {e}."
            )
            return fitted_specs[0]

    # ------------------------- Trial logic helpers. -------------------------
    @property
    def trials_from_node(self) -> set[int]:
        """Returns a set mapping a GenerationNode to the trials it generated.

        Returns:
            Set[int]: A set containing all the trials indices generated by this node.
        """
        trials_from_node = set()
        for _idx, trial in self.experiment.trials.items():
            for gr in trial.generator_runs:
                if (
                    gr._generation_node_name is not None
                    and gr._generation_node_name == self.node_name
                ):
                    trials_from_node.add(trial.index)
        return trials_from_node

    @property
    def node_that_generated_last_gr(self) -> str | None:
        """Returns the name of the node that generated the last generator run.

        Returns:
            str: The name of the node that generated the last generator run.
        """
        return (
            self.generation_strategy.last_generator_run._generation_node_name
            if self.generation_strategy.last_generator_run
            else None
        )

    @property
    def transition_edges(self) -> dict[str, list[TransitionCriterion]]:
        """Returns a dictionary mapping the next ``GenerationNode`` to the
        TransitionCriteria that define the transition that that node.

        Ex: if the transition from the current node to node `x` is defined by
        IsSingleObjective and MinTrials criterion then the return would be
        {'x': [IsSingleObjective, MinTrials]}.

        Returns:
            Dict[str, List[TransitionCriterion]]: A dictionary mapping the next
            ``GenerationNode`` to the ``TransitionCriterion`` that are associated
            with it.
        """
        if self.transition_criteria is None:
            return {}

        tc_edges = defaultdict(list)
        for tc in self.transition_criteria:
            tc_edges[tc.transition_to].append(tc)
        return tc_edges

    def should_transition_to_next_node(
        self, raise_data_required_error: bool = True
    ) -> tuple[bool, str]:
        """Checks whether we should transition to the next node based on this node's
        TransitionCriterion.

        Important: This method relies on the ``transition_criterion`` of this node to
        be listed in order of importance. Ex: a fallback transition should come after
        the primary transition in the transition criterion list.

        Args:
            raise_data_required_error: Whether to raise ``DataRequiredError`` in the
                case detailed above. Not raising the error is useful if just looking to
                check how many generator runs (to be made into trials) can be produced,
                but not actually producing them yet.
        Returns:
            Tuple[bool, str]: Whether we should transition to the next node
                and the name of the node to gen from (either the current or next node)
        """
        # if no transition criteria are defined, this node can generate unlimited trials
        if len(self.transition_criteria) == 0:
            return False, self.node_name

        # For each "transition edge" (set of all transition criteria that lead from
        # current node (e.g. "node A") to another specific node ("e.g. "node B")
        # in the node DAG:
        # I. Check if all of the transition criteria along that edge are met; if so,
        # transition to the next node defined by that edge.
        # II. If we did not transition along this edge, but the edge has some
        # "generation blocking" transition criteria (ex `MaxGenerationParallelism`)
        # that are met, raise the error associated with that criterion.
        for next_node, all_tc in self.transition_edges.items():
            # I. Check if there are any TCs that block transition and whether all
            # of them are met. If all fo them are met, then we should transition.
            transition_blocking = [tc for tc in all_tc if tc.block_transition_if_unmet]
            all_transition_blocking_met_should_transition = transition_blocking and all(
                tc.is_met(
                    experiment=self.experiment,
                    curr_node=self,
                )
                for tc in transition_blocking
            )
            if all_transition_blocking_met_should_transition:
                return True, next_node

            # II. Raise any necessary generation errors: for any met criterion,
            # call its `block_continued_generation_error` method if not all
            # transition-blocking criteria are met. The method might not raise an
            # error, depending on its implementation on given criterion, so the error
            # from the first met one that does block continued generation, will raise.
            if raise_data_required_error:
                generation_blocking = [tc for tc in all_tc if tc.block_gen_if_met]
                for tc in generation_blocking:
                    if tc.is_met(self.experiment, curr_node=self):
                        tc.block_continued_generation_error(
                            node_name=self.node_name,
                            experiment=self.experiment,
                            trials_from_node=self.trials_from_node,
                        )
                # TODO[@mgarrard, @drfreund] Try replacing `block_gen_if_met` with
                # a self-transition and rework this error block.

        return False, self.node_name

    def new_trial_limit(self, raise_generation_errors: bool = False) -> int:
        """How many trials can this generation strategy can currently produce
        ``GeneratorRun``-s for (with potentially multiple generator runs produced for
        each intended trial).

        NOTE: Only considers transition criteria that inherit from
        ``TrialBasedCriterion``.

        Returns:
            The number of generator runs that can currently be produced, with -1
            meaning unlimited generator runs.
        """
        # TODO: @mgarrard Should we consider returning `None` if there is no limit?
        trial_based_gen_blocking_criteria = [
            criterion
            for criterion in self.transition_criteria
            if criterion.block_gen_if_met and isinstance(criterion, TrialBasedCriterion)
        ]
        gen_blocking_criterion_delta_from_threshold = [
            criterion.num_till_threshold(
                experiment=self.experiment, trials_from_node=self.trials_from_node
            )
            for criterion in trial_based_gen_blocking_criteria
        ]

        # Raise any necessary generation errors: for any met criterion,
        # call its `block_continued_generation_error` method The method might not
        # raise an error, depending on its implementation on given criterion, so the
        # error from the first met one that does block continued generation, will be
        # raised.
        if raise_generation_errors:
            for criterion in trial_based_gen_blocking_criteria:
                # TODO[mgarrard]: Raise a group of all the errors, from each gen-
                # blocking transition criterion.
                if criterion.is_met(
                    self.experiment,
                    curr_node=self,
                ):
                    criterion.block_continued_generation_error(
                        node_name=self.node_name,
                        experiment=self.experiment,
                        trials_from_node=self.trials_from_node,
                    )
        if len(gen_blocking_criterion_delta_from_threshold) == 0:
            return -1
        return min(gen_blocking_criterion_delta_from_threshold)

    def _brief_transition_criteria_repr(self) -> str:
        """Returns a brief string representation of the
        transition criteria for this node.

        Returns:
            str: A string representation of the transition criteria for this node.
        """
        if self.transition_criteria is None:
            return "None"
        tc_list = ", ".join(
            [
                f"{tc.__class__.__name__}(transition_to='{str(tc.transition_to)}')"
                for tc in self.transition_criteria
            ]
        )
        return f"[{tc_list}]"

    def apply_input_constructors(
        self,
        experiment: Experiment,
        gen_kwargs: dict[str, Any],
    ) -> dict[str, Union[int, ObservationFeatures | None]]:
        # NOTE: In the future we might have to add new types ot the `Union` above
        # or allow `Any` for the value type, but until we have more different types
        # of input constructors, this provides a bit of additional typechecking.
        return {
            "n": self._determine_arms_from_node(
                experiment=experiment,
                gen_kwargs=gen_kwargs,
            ),
            "fixed_features": self._determine_fixed_features_from_node(
                experiment=experiment,
                gen_kwargs=gen_kwargs,
            ),
        }

    def _determine_arms_from_node(
        self,
        experiment: Experiment,
        gen_kwargs: dict[str, Any],
    ) -> int:
        """Calculates the number of arms to generate from the node that will be used
        during generation.

        Args:
            gen_kwargs: The kwargs passed to the ``GenerationStrategy``'s
                gen call, including arms_per_node: an optional map from node name to
                the number of arms to generate from that node. If not provided, will
                default to the number of arms specified in the node's
                ``InputConstructors`` or n if no``InputConstructors`` are defined on
                the node.

        Returns:
            The number of arms to generate from the node that will be used during this
            generation via ``_gen_multiple``.
        """
        arms_per_node = gen_kwargs.get("arms_per_node")
        purpose_N = (
            gs_module.generation_node_input_constructors.InputConstructorPurpose.N
        )
        if arms_per_node is not None:
            # arms_per_node provides a way to manually override input
            # constructors. This should be used with caution, and only
            # if you really know what you're doing. :)
            arms_from_node = arms_per_node[self.node_name]
        elif purpose_N not in self.input_constructors:
            # if the node does not have an input constructor for N, we first
            # use the n passed to method if it exists, then check if the model gen
            # kwargs define an n, then fallback to default values of n
            arms_from_node = gen_kwargs.get("n")
            if arms_from_node is None and self.generator_spec_to_gen_from is not None:
                arms_from_node = self.generator_spec_to_gen_from.model_gen_kwargs.get(
                    "n", None
                )
            if arms_from_node is None:
                # TODO[@mgarrard, @drfreund]: We can remove this check if we
                # decide that generation nodes can only be used within a
                # generation strategy; then we will need to refactor some tests.
                if self._generation_strategy is not None:
                    arms_from_node = self._generation_strategy.DEFAULT_N
                else:
                    arms_from_node = 1
        else:
            arms_from_node = self.input_constructors[purpose_N](
                previous_node=self.previous_node,
                next_node=self,
                gs_gen_call_kwargs=gen_kwargs,
                experiment=experiment,
            )

        return arms_from_node

    def _determine_fixed_features_from_node(
        self,
        experiment: Experiment,
        gen_kwargs: dict[str, Any],
    ) -> ObservationFeatures | None:
        """Uses the ``InputConstructors`` on the node to determine the fixed features
        to pass into the model. If fixed_features are provided, they will take
        precedence over the fixed_features from the node.

        Args:
            node_to_gen_from: The node from which to generate from
            gen_kwargs: The kwargs passed to the ``GenerationStrategy``'s
                gen call, including the fixed features passed to the ``gen`` method if
                any.

        Returns:
            An object of ObservationFeatures that represents the fixed features to
            pass into the model.
        """
        # passed_fixed_features represents the fixed features that were passed by the
        # user to the gen method as overrides.
        passed_fixed_features = gen_kwargs.get("fixed_features")
        if passed_fixed_features is not None:
            return passed_fixed_features

        node_fixed_features = None
        input_constructors_module = gs_module.generation_node_input_constructors
        purpose_fixed_features = (
            input_constructors_module.InputConstructorPurpose.FIXED_FEATURES
        )
        if purpose_fixed_features in self.input_constructors:
            node_fixed_features = self.input_constructors[purpose_fixed_features](
                previous_node=self.previous_node,
                next_node=self,
                gs_gen_call_kwargs=gen_kwargs,
                experiment=experiment,
            )
        return node_fixed_features


class GenerationStep(GenerationNode, SortableBase):
    """One step in the generation strategy, corresponds to a single generator.
    Describes the generator, how many trials will be generated with this generator, what
    minimum number of observations is required to proceed to the next generator, etc.

    Args:
        generator: A member of `Generators` enum returning an instance of
            `Adapter` with an instantiated underlying `Generator`.
        num_trials: How many trials to generate with the model from this step.
            If set to -1, trials will continue to be generated from this model
            as long as `generation_strategy.gen` is called (available only for
            the last of the generation steps).
        min_trials_observed: How many trials must be completed before the
            generation strategy can proceed to the next step. Defaults to 0.
            If `num_trials` of a given step have been generated but `min_trials_
            observed` have not been completed, a call to `generation_strategy.gen`
            will fail with a `DataRequiredError`.
        max_parallelism: How many trials generated in the course of this step are
            allowed to be run (i.e. have `trial.status` of `RUNNING`) simultaneously.
            If `max_parallelism` trials from this step are already running, a call
            to `generation_strategy.gen` will fail with a `MaxParallelismReached
            Exception`, indicating that more trials need to be completed before
            generating and running next trials.
        use_update: DEPRECATED.
        enforce_num_trials: Whether to enforce that only `num_trials` are generated
            from the given step. If False and `num_trials` have been generated, but
            `min_trials_observed` have not been completed, `generation_strategy.gen`
            will continue generating trials from the current step, exceeding `num_
            trials` for it. Allows to avoid `DataRequiredError`, but delays
            proceeding to next generation step.
        model_kwargs: Dictionary of kwargs to pass into the adapter and generator
            constructors on instantiation. E.g. if `generator` is `Generators.SOBOL`,
            kwargs will be applied as `Generators.SOBOL(**model_kwargs)`.
            NOTE: If generation strategy is interrupted and resumed from a stored
            snapshot and its last used generator has state saved on its generator runs,
            `model_kwargs` is updated with the state dict of the generator, retrieved
            from the last generator run of this generation strategy.
        model_gen_kwargs: Each call to `generation_strategy.gen` performs a call
            to the step's adapter's `gen` under the hood; `model_gen_kwargs` will be
            passed to the adapter's `gen` like: `adapter.gen(**model_gen_kwargs)`.
        completion_criteria: List of TransitionCriterion. All `is_met` must evaluate
            True for the GenerationStrategy to move on to the next Step
        index: Index of this generation step, for use internally in `Generation
            Strategy`. Do not assign as it will be reassigned when instantiating
            `GenerationStrategy` with a list of its steps.
        should_deduplicate: Whether to deduplicate the parameters of proposed arms
            against those of previous arms via rejection sampling. If this is True,
            the generation strategy will discard generator runs produced from the
            generation step that has `should_deduplicate=True` if they contain arms
            already present on the experiment and replace them with new generator runs.
            If no generator run with entirely unique arms could be produced in 5
            attempts, a `GenerationStrategyRepeatedPoints` error will be raised, as we
            assume that the optimization converged when the model can no longer suggest
            unique arms.
        generator_name: Optional name of the generator. If not specified, defaults to
            the `model_key` of the generator spec.

    Note for developers: by "generator" here we really mean an ``Adapter`` object, which
    contains a ``Generator`` under the hood. We call it "generator" here to simplify and
    focus on explaining the logic of GenerationStep and GenerationStrategy.
    """

    def __init__(
        self,
        generator: GeneratorRegistryBase,
        num_trials: int,
        model_kwargs: dict[str, Any] | None = None,
        model_gen_kwargs: dict[str, Any] | None = None,
        completion_criteria: Sequence[TransitionCriterion] | None = None,
        min_trials_observed: int = 0,
        max_parallelism: int | None = None,
        enforce_num_trials: bool = True,
        should_deduplicate: bool = False,
        generator_name: str | None = None,
        use_update: bool = False,  # DEPRECATED.
        index: int = -1,  # Index of this step, set internally.
    ) -> None:
        r"""Initializes a single-model GenerationNode, a.k.a. a GenerationStep.

        See the class docstring for argument descriptions.
        """
        if use_update:
            raise DeprecationWarning("`GenerationStep.use_update` is deprecated.")
        # These are here for backwards compatibility. Prior to implementation of
        # the __init__ method, these were the fields of the dataclass. GenerationStep
        # storage utilizes these attributes, so we need to store them. Once we start
        # using GenerationNode storage, we can clean up these attributes.
        self.index = index
        self.generator = generator
        self.num_trials = num_trials
        self.completion_criteria: Sequence[TransitionCriterion] = (
            completion_criteria or []
        )
        self.min_trials_observed = min_trials_observed
        self.max_parallelism = max_parallelism
        self.enforce_num_trials = enforce_num_trials
        self.use_update = use_update

        model_kwargs = model_kwargs or {}
        model_gen_kwargs = model_gen_kwargs or {}

        if (
            self.enforce_num_trials
            and (self.num_trials >= 0)
            and (self.min_trials_observed > self.num_trials)
        ):
            raise UserInputError(
                "`GenerationStep` received `min_trials_observed > num_trials` "
                f"(`min_trials_observed = {self.min_trials_observed}`, `num_trials = "
                f"{self.num_trials}`), making completion of this step impossible. "
                "Please alter inputs so that `min_trials_observed <= num_trials`."
            )
        if not isinstance(self.generator, GeneratorRegistryBase):
            raise UserInputError(
                "`generator` in generation step must be a `GeneratorRegistryBase` "
                "enum subclass entry. Support for alternative callables "
                f"has been deprecated. Got {self.generator=}."
            )
        else:
            generator_spec = GeneratorSpec(
                generator_enum=self.generator,
                model_kwargs=model_kwargs,
                model_gen_kwargs=model_gen_kwargs,
            )
        if not generator_name:
            generator_name = generator_spec.model_key
        self.generator_name: str = generator_name

        # Create transition criteria for this step. If num_trials is provided to
        # this `GenerationStep`, then we create a `MinTrials` criterion which ensures
        # at least that many trials in good status are generated. `MinTrials` can also
        # enforce the min_trials_observed requirement. The `transition_to` argument
        # is set in `GenerationStrategy` constructor, because only then is the order
        # of the generation steps actually known.
        transition_criteria = []
        if self.num_trials != -1:
            transition_criteria.append(
                MinTrials(
                    threshold=self.num_trials,
                    not_in_statuses=[TrialStatus.FAILED, TrialStatus.ABANDONED],
                    block_gen_if_met=self.enforce_num_trials,
                    block_transition_if_unmet=True,
                )
            )

        if self.min_trials_observed > 0:
            transition_criteria.append(
                MinTrials(
                    only_in_statuses=[
                        TrialStatus.COMPLETED,
                        TrialStatus.EARLY_STOPPED,
                    ],
                    threshold=self.min_trials_observed,
                    block_gen_if_met=False,
                    block_transition_if_unmet=True,
                )
            )
        max_parallelism = self.max_parallelism
        if max_parallelism is not None:
            transition_criteria.append(
                MaxGenerationParallelism(
                    threshold=max_parallelism,
                    only_in_statuses=[TrialStatus.RUNNING],
                    block_gen_if_met=True,
                    block_transition_if_unmet=False,
                    transition_to=None,
                )
            )

        transition_criteria += self.completion_criteria
        super().__init__(
            node_name=f"GenerationStep_{str(self.index)}",
            generator_specs=[generator_spec],
            should_deduplicate=should_deduplicate,
            transition_criteria=transition_criteria,
        )

    @property
    def model_kwargs(self) -> dict[str, Any]:
        """Returns the model kwargs of the underlying ``GeneratorSpec``."""
        return self.generator_spec.model_kwargs

    @property
    def model_gen_kwargs(self) -> dict[str, Any]:
        """Returns the model gen kwargs of the underlying ``GeneratorSpec``."""
        return self.generator_spec.model_gen_kwargs

    @property
    def generator_spec(self) -> GeneratorSpec:
        """Returns the first generator_spec from the generator_specs attribute."""
        return self.generator_specs[0]

    @property
    def _unique_id(self) -> str:
        """Returns the unique ID of this generation step, which is the index."""
        return str(self.index)

    def gen(
        self,
        *,
        experiment: Experiment,
        n: int | None = None,
        pending_observations: dict[str, list[ObservationFeatures]] | None,
        skip_fit: bool = False,
        data: Data | None = None,
        **gs_gen_kwargs: Any,
    ) -> GeneratorRun | None:
        gr = super().gen(
            experiment=experiment,
            n=n,
            pending_observations=pending_observations,
            data=data,
            skip_fit=skip_fit,
            **gs_gen_kwargs,
        )
        if gr is None:
            raise AxError(
                "This `GenerationStep`'s underlying `GenerationNode` returned "
                "`None` from `gen`. This is an unexpected state for a "
                "`GenerationStep`-based generation strategy."
            )
        gr._generation_step_index = self.index
        return gr



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from contextlib import ExitStack
from unittest.mock import patch

import numpy as np
import torch
from ax.adapter.adapter_utils import (
    get_pareto_frontier_and_configs,
    observed_hypervolume,
    observed_pareto_frontier,
    pareto_frontier,
    predicted_hypervolume,
    predicted_pareto_frontier,
)
from ax.adapter.factory import get_sobol
from ax.adapter.registry import Cont_X_trans, ST_MTGP_trans, Y_trans
from ax.adapter.torch import TorchAdapter
from ax.core.metric import Metric
from ax.core.observation import ObservationData, ObservationFeatures
from ax.core.optimization_config import MultiObjectiveOptimizationConfig
from ax.core.outcome_constraint import (
    ComparisonOp,
    ObjectiveThreshold,
    OutcomeConstraint,
)
from ax.core.parameter_constraint import ParameterConstraint
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.generators.torch.botorch_moo import MultiObjectiveLegacyBoTorchGenerator
from ax.generators.torch.botorch_moo_defaults import (
    infer_objective_thresholds,
    pareto_frontier_evaluator,
)
from ax.utils.common.random import set_rng_seed
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_data_multi_objective,
    get_branin_experiment_with_multi_objective,
    get_hierarchical_search_space,
    get_hss_trials_with_fixed_parameter,
    TEST_SOBOL_SEED,
)
from ax.utils.testing.mock import mock_botorch_optimize, skip_fit_gpytorch_mll
from botorch.utils.multi_objective.pareto import is_non_dominated
from pyre_extensions import assert_is_instance, none_throws

PARETO_FRONTIER_EVALUATOR_PATH = (
    f"{get_pareto_frontier_and_configs.__module__}.pareto_frontier_evaluator"
)
STUBS_PATH: str = get_branin_experiment_with_multi_objective.__module__


class MultiObjectiveTorchAdapterTest(TestCase):
    @patch(
        # Mocking `BraninMetric` as not available while running, so it will
        # be grabbed from cache during `fetch_data`.
        f"{STUBS_PATH}.BraninMetric.is_available_while_running",
        return_value=False,
    )
    @skip_fit_gpytorch_mll
    def helper_test_pareto_frontier(
        self, _, outcome_constraints: list[OutcomeConstraint] | None
    ) -> None:
        """
        Make sure Pareto-related functions run.

        Data is generated manually; this does not check that the points are actually
        Pareto-efficient or that everything works end-to-end.
        """
        exp = get_branin_experiment_with_multi_objective(
            has_optimization_config=True, with_batch=True
        )
        for trial in exp.trials.values():
            trial.mark_running(no_runner_required=True).mark_completed()
        metrics_dict = none_throws(exp.optimization_config).metrics
        objective_bound = 5.0
        objective_thresholds = [
            ObjectiveThreshold(
                metric=metrics_dict[f"branin_{letter}"],
                bound=objective_bound,
                relative=False,
                op=ComparisonOp.LEQ,
            )
            for letter in "ab"
        ]
        exp.optimization_config = assert_is_instance(
            exp.optimization_config, MultiObjectiveOptimizationConfig
        ).clone_with_args(
            objective_thresholds=objective_thresholds,
            outcome_constraints=outcome_constraints,
        )

        n_outcomes = len(exp.metrics)
        exp.attach_data(
            get_branin_data_multi_objective(
                trial_indices=exp.trials.keys(), outcomes=list(exp.metrics)
            ),
        )
        adapter = TorchAdapter(
            experiment=exp,
            generator=MultiObjectiveLegacyBoTorchGenerator(),
        )
        with patch(
            PARETO_FRONTIER_EVALUATOR_PATH, wraps=pareto_frontier_evaluator
        ) as wrapped_frontier_evaluator:
            observed_frontier = observed_pareto_frontier(
                adapter=adapter, objective_thresholds=objective_thresholds
            )

        wrapped_frontier_evaluator.assert_called_once()
        self.assertIsNone(wrapped_frontier_evaluator.call_args[1]["X"])
        self.assertEqual(1, len(observed_frontier))
        self.assertEqual(observed_frontier[0].arm_name, "0_0")

        with self.assertRaises(ValueError):
            predicted_pareto_frontier(
                adapter=adapter,
                objective_thresholds=objective_thresholds,
                observation_features=[],
            )

        predicted_frontier = predicted_pareto_frontier(
            adapter=adapter,
            objective_thresholds=objective_thresholds,
            observation_features=None,
        )
        self.assertEqual(predicted_frontier[0].arm_name, "0_0")

        observation_features = [
            ObservationFeatures(parameters={"x1": 0.0, "x2": 1.0}),
            # magic number to ensure the model makes one non-negative
            # prediction, so there is a point within the thresholds
            ObservationFeatures(parameters={"x1": -2.0, "x2": 6.3}),
        ]
        extra_outcome = ["branin_c"] if outcome_constraints is not None else []
        observation_data = [
            ObservationData(
                metric_names=["branin_b", "branin_a"] + extra_outcome,
                means=np.arange(1, 1 + n_outcomes),
                covariance=np.eye(n_outcomes),
            ),
            ObservationData(
                metric_names=["branin_a", "branin_b"] + extra_outcome,
                means=np.arange(3, 3 + n_outcomes),
                covariance=np.eye(n_outcomes),
            ),
        ]
        # appease Pyre (this condition is True)
        if isinstance(exp.optimization_config, MultiObjectiveOptimizationConfig):
            predicted_frontier = predicted_pareto_frontier(
                adapter=adapter,
                objective_thresholds=objective_thresholds,
                observation_features=observation_features,
                optimization_config=exp.optimization_config,
            )
        self.assertTrue(len(predicted_frontier) <= 2)
        self.assertIsNone(predicted_frontier[0].arm_name, None)

        with patch(
            PARETO_FRONTIER_EVALUATOR_PATH, wraps=pareto_frontier_evaluator
        ) as wrapped_frontier_evaluator:
            (
                observed_frontier,
                f,
                obj_w,
                obj_t,
            ) = get_pareto_frontier_and_configs(
                adapter=adapter,
                objective_thresholds=objective_thresholds,
                observation_features=observation_features,
                observation_data=observation_data,
                use_model_predictions=False,
            )

        wrapped_frontier_evaluator.assert_called_once()
        self.assertEqual(f.shape, (1, n_outcomes))
        self.assertTrue(torch.equal(obj_w[:2], -torch.ones(2, dtype=torch.double)))
        self.assertTrue(obj_t is not None)
        self.assertTrue(
            torch.equal(
                none_throws(obj_t)[:2], torch.full((2,), 5.0, dtype=torch.double)
            )
        )
        observed_frontier2 = pareto_frontier(
            adapter=adapter,
            objective_thresholds=objective_thresholds,
            observation_features=observation_features,
            observation_data=observation_data,
            use_model_predictions=False,
        )
        self.assertEqual(observed_frontier, observed_frontier2)

        # Remove the thresholds for testing None handling.
        assert_is_instance(
            adapter._optimization_config, MultiObjectiveOptimizationConfig
        )._objective_thresholds = []
        with patch(
            PARETO_FRONTIER_EVALUATOR_PATH, wraps=pareto_frontier_evaluator
        ) as wrapped_frontier_evaluator:
            (
                observed_frontier,
                f,
                obj_w,
                obj_t,
            ) = get_pareto_frontier_and_configs(
                adapter=adapter,
                objective_thresholds=None,
                observation_features=observation_features,
                observation_data=observation_data,
                use_model_predictions=False,
            )
        wrapped_frontier_evaluator.assert_called_once()
        self.assertIsNone(wrapped_frontier_evaluator.call_args[1]["X"])
        true_Y = torch.tensor([[2.0, 1.0], [3.0, 4.0]], dtype=torch.double)
        self.assertTrue(
            torch.equal(
                wrapped_frontier_evaluator.call_args[1]["Y"][:, :2],
                true_Y,
            )
        )
        self.assertTrue(torch.equal(objective_bound - f[:, :2], true_Y[1:, :]))
        self.assertIsNone(obj_t)

    def test_pareto_frontier(self) -> None:
        """
        Run helper_test_pareto_frontier with and without outcome constraints.

        The constraint won't come close to binding, so it shouldn't affect results.
        """
        constraint = OutcomeConstraint(
            Metric(name="branin_c"), ComparisonOp.LEQ, bound=100.0, relative=False
        )
        for outcome_constraints in [None, [constraint]]:
            with self.subTest(outcome_constraints=outcome_constraints):
                self.helper_test_pareto_frontier(
                    outcome_constraints=outcome_constraints
                )

    @mock_botorch_optimize
    def test_get_pareto_frontier_and_configs_input_validation(self) -> None:
        exp = get_branin_experiment_with_multi_objective(
            has_optimization_config=True, with_batch=True
        )
        for trial in exp.trials.values():
            trial.mark_running(no_runner_required=True).mark_completed()
        metrics_dict = none_throws(exp.optimization_config).metrics
        objective_thresholds = [
            ObjectiveThreshold(
                metric=metrics_dict[f"branin_{letter}"],
                bound=5.0,
                relative=False,
                op=ComparisonOp.LEQ,
            )
            for letter in "ab"
        ]
        exp.optimization_config = assert_is_instance(
            exp.optimization_config, MultiObjectiveOptimizationConfig
        ).clone_with_args(
            objective_thresholds=objective_thresholds,
        )

        exp.attach_data(
            get_branin_data_multi_objective(trial_indices=exp.trials.keys()),
        )
        adapter = TorchAdapter(
            experiment=exp,
            search_space=exp.search_space,
            data=exp.fetch_data(),
            generator=MultiObjectiveLegacyBoTorchGenerator(),
            transforms=[],
        )
        observation_features = [
            ObservationFeatures(parameters={"x1": 0.0, "x2": 1.0}),
            ObservationFeatures(parameters={"x1": 1.0, "x2": 0.0}),
        ]

        with self.assertWarns(
            Warning,
            msg="You provided `observation_data` when `use_model_predictions` is True; "
            "`observation_data` will not be used.",
        ):
            res = get_pareto_frontier_and_configs(
                adapter,
                observation_features=observation_features,
                use_model_predictions=True,
                observation_data=[],
            )
            self.assertEqual(len(res), 4)

        with self.assertRaises(ValueError):
            get_pareto_frontier_and_configs(
                adapter, observation_features=[], use_model_predictions=False
            )

    @patch(
        # Mocking `BraninMetric` as not available while running, so it will
        # be grabbed from cache during `fetch_data`.
        f"{STUBS_PATH}.BraninMetric.is_available_while_running",
        return_value=False,
    )
    def test_hypervolume(self, _, cuda: bool = False) -> None:
        for num_objectives in (2, 3):
            exp = get_branin_experiment_with_multi_objective(
                has_optimization_config=True,
                with_batch=True,
                num_objectives=num_objectives,
            )
            for trial in exp.trials.values():
                trial.mark_running(no_runner_required=True).mark_completed()
            # pyre-fixme[16]: Optional type has no attribute `metrics`.
            metrics_dict = exp.optimization_config.metrics
            # Objective thresholds and synthetic observations chosen to have closed-form
            # hypervolumes to test.
            objective_thresholds = [
                ObjectiveThreshold(
                    metric=metrics_dict["branin_a"],
                    bound=10.0,
                    relative=False,
                    op=ComparisonOp.LEQ,
                ),
                ObjectiveThreshold(
                    metric=metrics_dict["branin_b"],
                    bound=9.0,
                    relative=False,
                    op=ComparisonOp.LEQ,
                ),
            ]
            if num_objectives == 3:
                objective_thresholds.append(
                    ObjectiveThreshold(
                        metric=metrics_dict["branin_c"],
                        bound=8.0,
                        relative=False,
                        op=ComparisonOp.LEQ,
                    )
                )
            optimization_config = assert_is_instance(
                exp.optimization_config, MultiObjectiveOptimizationConfig
            ).clone_with_args(objective_thresholds=objective_thresholds)
            exp.attach_data(
                get_branin_data_multi_objective(
                    trial_indices=exp.trials.keys(),
                    outcomes=list(optimization_config.metrics),
                )
            )
            adapter = TorchAdapter(
                search_space=exp.search_space,
                generator=MultiObjectiveLegacyBoTorchGenerator(),
                optimization_config=optimization_config,
                transforms=[],
                experiment=exp,
                data=exp.fetch_data(),
                torch_device=torch.device("cuda" if cuda else "cpu"),
            )
            with patch(
                PARETO_FRONTIER_EVALUATOR_PATH, wraps=pareto_frontier_evaluator
            ) as wrapped_frontier_evaluator:
                hv = observed_hypervolume(
                    adapter=adapter, objective_thresholds=objective_thresholds
                )
            expected_hv = 20 if num_objectives == 2 else 60  # 5 * 4 (* 3)
            wrapped_frontier_evaluator.assert_called_once()
            self.assertEqual(expected_hv, hv)
            if num_objectives == 3:
                # Test selected_metrics
                hv = observed_hypervolume(
                    adapter=adapter,
                    objective_thresholds=objective_thresholds,
                    selected_metrics=["branin_a", "branin_c"],
                )
                expected_hv = 15  # (5 - 0) * (5 - 2)
                self.assertEqual(expected_hv, hv)
                # test that non-objective outcome raises value error
                with self.assertRaises(ValueError):
                    hv = observed_hypervolume(
                        adapter=adapter,
                        objective_thresholds=objective_thresholds,
                        selected_metrics=["tracking"],
                    )

            with self.assertRaises(ValueError):
                predicted_hypervolume(
                    adapter=adapter,
                    objective_thresholds=objective_thresholds,
                    observation_features=[],
                )

            observation_features = [
                ObservationFeatures(parameters={"x1": 1.0, "x2": 2.0}),
                ObservationFeatures(parameters={"x1": 2.0, "x2": 1.0}),
            ]
            predicted_hv = predicted_hypervolume(
                adapter=adapter,
                objective_thresholds=objective_thresholds,
                observation_features=observation_features,
            )
            self.assertTrue(predicted_hv >= 0)
            if num_objectives == 3:
                # Test selected_metrics
                predicted_hv = predicted_hypervolume(
                    adapter=adapter,
                    objective_thresholds=objective_thresholds,
                    observation_features=observation_features,
                    selected_metrics=["branin_a", "branin_c"],
                )
                self.assertTrue(predicted_hv >= 0)

    def test_hypervolume_cuda(self) -> None:
        if torch.cuda.is_available():
            self.test_hypervolume(cuda=True)

    @patch(
        # Mocking `BraninMetric` as not available while running, so it will
        # be grabbed from cache during `fetch_data`.
        f"{STUBS_PATH}.BraninMetric.is_available_while_running",
        return_value=False,
    )
    @mock_botorch_optimize
    def test_infer_objective_thresholds(self, _, cuda: bool = False) -> None:
        # lightweight test
        exp = get_branin_experiment_with_multi_objective(
            has_optimization_config=True,
            with_batch=True,
            with_status_quo=True,
        )
        for trial in exp.trials.values():
            trial.mark_running(no_runner_required=True).mark_completed()
        exp.attach_data(
            get_branin_data_multi_objective(trial_indices=exp.trials.keys())
        )
        data = exp.fetch_data()
        adapter = TorchAdapter(
            search_space=exp.search_space,
            generator=MultiObjectiveLegacyBoTorchGenerator(),
            optimization_config=exp.optimization_config,
            transforms=Cont_X_trans + Y_trans,
            torch_device=torch.device("cuda" if cuda else "cpu"),
            experiment=exp,
            data=data,
        )
        fixed_features = ObservationFeatures(parameters={"x1": 0.0})
        search_space = exp.search_space.clone()
        param_constraints = [
            ParameterConstraint(constraint_dict={"x1": 1.0}, bound=10.0)
        ]
        search_space.add_parameter_constraints(param_constraints)
        oc = none_throws(exp.optimization_config).clone()
        oc.objective._objectives[0].minimize = True

        for use_partial_thresholds in (False, True):
            if use_partial_thresholds:
                assert_is_instance(
                    oc, MultiObjectiveOptimizationConfig
                )._objective_thresholds = [
                    ObjectiveThreshold(
                        metric=oc.objective.metrics[0],
                        bound=2.0,
                        relative=False,
                        op=ComparisonOp.LEQ,
                    )
                ]
            expected_base_gen_args = adapter._get_transformed_gen_args(
                search_space=search_space.clone(),
                optimization_config=oc,
                fixed_features=fixed_features,
            )
            with ExitStack() as es:
                mock_model_infer_obj_t = es.enter_context(
                    patch(
                        "ax.adapter.torch.infer_objective_thresholds",
                        wraps=infer_objective_thresholds,
                    )
                )
                mock_get_transformed_gen_args = es.enter_context(
                    patch.object(
                        adapter,
                        "_get_transformed_gen_args",
                        wraps=adapter._get_transformed_gen_args,
                    )
                )
                mock_get_transformed_model_gen_args = es.enter_context(
                    patch.object(
                        adapter,
                        "_get_transformed_model_gen_args",
                        wraps=adapter._get_transformed_model_gen_args,
                    )
                )
                mock_untransform_objective_thresholds = es.enter_context(
                    patch.object(
                        adapter,
                        "_untransform_objective_thresholds",
                        wraps=adapter._untransform_objective_thresholds,
                    )
                )
                obj_thresholds = adapter.infer_objective_thresholds(
                    search_space=search_space,
                    optimization_config=oc,
                    fixed_features=fixed_features,
                )
                expected_obj_weights = torch.tensor([-1.0, -1.0], dtype=torch.double)
                ckwargs = mock_model_infer_obj_t.call_args[1]
                self.assertTrue(
                    torch.equal(ckwargs["objective_weights"], expected_obj_weights)
                )
                # check that transforms have been applied (at least UnitX)
                self.assertEqual(ckwargs["bounds"], [(0.0, 1.0), (0.0, 1.0)])
                lc = ckwargs["linear_constraints"]
                self.assertTrue(
                    torch.equal(lc[0], torch.tensor([[15.0, 0.0]], dtype=torch.double))
                )
                self.assertTrue(
                    torch.equal(lc[1], torch.tensor([[15.0]], dtype=torch.double))
                )
                self.assertEqual(ckwargs["fixed_features"], {0: 1.0 / 3.0})
                mock_get_transformed_gen_args.assert_called_once()
                mock_get_transformed_model_gen_args.assert_called_once_with(
                    search_space=expected_base_gen_args.search_space,
                    fixed_features=expected_base_gen_args.fixed_features,
                    pending_observations=expected_base_gen_args.pending_observations,
                    optimization_config=expected_base_gen_args.optimization_config,
                )
                mock_untransform_objective_thresholds.assert_called_once()
                ckwargs = mock_untransform_objective_thresholds.call_args[1]

                self.assertTrue(
                    torch.equal(ckwargs["objective_weights"], expected_obj_weights)
                )
            self.assertEqual(obj_thresholds[0].metric.name, "branin_a")
            self.assertEqual(obj_thresholds[1].metric.name, "branin_b")
            self.assertEqual(obj_thresholds[0].op, ComparisonOp.LEQ)
            self.assertEqual(obj_thresholds[1].op, ComparisonOp.LEQ)
            self.assertFalse(obj_thresholds[0].relative)
            self.assertFalse(obj_thresholds[1].relative)
            df = exp.to_df()
            Y = np.stack([df.branin_a.values, df.branin_b.values]).T
            Y = torch.from_numpy(Y)
            Y[:, 0] *= -1
            pareto_Y = Y[is_non_dominated(Y)]
            nadir = pareto_Y.min(dim=0).values
            self.assertTrue(
                np.all(
                    -np.array([obj_thresholds[0].bound, obj_thresholds[1].bound])
                    < nadir.numpy()
                )
            )

        # test using MTGP
        sobol_generator = get_sobol(
            search_space=exp.search_space,
            seed=TEST_SOBOL_SEED,
            # set initial position equal to the number of sobol arms generated
            # so far. This means that new sobol arms will complement the previous
            # arms in a space-filling fashion
            init_position=len(exp.arms_by_name) - 1,
        )
        sobol_run = sobol_generator.gen(n=2)
        trial = exp.new_batch_trial(should_add_status_quo_arm=True)
        trial.add_generator_run(sobol_run)
        trial.mark_running(no_runner_required=True).mark_completed()
        data = exp.fetch_data()
        set_rng_seed(0)  # make model fitting deterministic
        adapter = TorchAdapter(
            search_space=exp.search_space,
            generator=MultiObjectiveLegacyBoTorchGenerator(),
            optimization_config=exp.optimization_config,
            transforms=ST_MTGP_trans,
            experiment=exp,
            data=data,
        )
        fixed_features = ObservationFeatures(parameters={}, trial_index=1)
        expected_base_gen_args = adapter._get_transformed_gen_args(
            search_space=search_space.clone(),
            optimization_config=exp.optimization_config,
            fixed_features=fixed_features,
        )
        with ExitStack() as es:
            mock_model_infer_obj_t = es.enter_context(
                patch(
                    "ax.adapter.torch.infer_objective_thresholds",
                    wraps=infer_objective_thresholds,
                )
            )
            mock_untransform_objective_thresholds = es.enter_context(
                patch.object(
                    adapter,
                    "_untransform_objective_thresholds",
                    wraps=adapter._untransform_objective_thresholds,
                )
            )
            obj_thresholds = adapter.infer_objective_thresholds(
                search_space=search_space,
                optimization_config=exp.optimization_config,
                fixed_features=fixed_features,
            )
            ckwargs = mock_model_infer_obj_t.call_args[1]
            self.assertEqual(ckwargs["fixed_features"], {2: 1.0})
            mock_untransform_objective_thresholds.assert_called_once()
        self.assertEqual(obj_thresholds[0].metric.name, "branin_a")
        self.assertEqual(obj_thresholds[1].metric.name, "branin_b")
        self.assertEqual(obj_thresholds[0].op, ComparisonOp.LEQ)
        self.assertEqual(obj_thresholds[1].op, ComparisonOp.LEQ)
        self.assertFalse(obj_thresholds[0].relative)
        self.assertFalse(obj_thresholds[1].relative)
        df = exp.to_df()
        trial_mask = df.trial_index == 1
        Y = np.stack([df.branin_a.values[trial_mask], df.branin_b.values[trial_mask]]).T
        Y = torch.from_numpy(Y)
        pareto_Y = Y[is_non_dominated(Y)]
        nadir = pareto_Y.min(dim=0).values
        self.assertTrue(
            np.all(
                np.array([obj_thresholds[0].bound, obj_thresholds[1].bound])
                < nadir.numpy()
            )
        )

        # test with HSS
        hss = get_hierarchical_search_space(with_fixed_parameter=True)
        exp = get_branin_experiment_with_multi_objective(has_optimization_config=True)
        data = get_branin_data_multi_objective(trial_indices=[0, 1])
        # Update trials to match the search space.
        exp._search_space = hss
        exp._trials = get_hss_trials_with_fixed_parameter(exp=exp)
        adapter = TorchAdapter(
            search_space=hss,
            generator=MultiObjectiveLegacyBoTorchGenerator(),
            optimization_config=exp.optimization_config,
            transforms=Cont_X_trans + Y_trans,
            torch_device=torch.device("cuda" if cuda else "cpu"),
            experiment=exp,
            data=data,
            # [T143911996] The trials get ignored without fit_out_of_design.
            fit_out_of_design=True,
        )
        self.assertIn("Cast", adapter.transforms)
        with patch.object(
            adapter,
            "_untransform_objective_thresholds",
            wraps=adapter._untransform_objective_thresholds,
        ) as mock_untransform, patch.object(
            adapter.transforms["Cast"],
            "untransform_observation_features",
            wraps=adapter.transforms["Cast"].untransform_observation_features,
        ) as wrapped_cast:
            obj_thresholds = adapter.infer_objective_thresholds(
                search_space=hss,
                optimization_config=exp.optimization_config,
                fixed_features=None,
            )
        mock_untransform.assert_called_once()
        self.assertEqual(wrapped_cast.call_count, 0)
        self.assertEqual(obj_thresholds[0].metric.name, "branin_a")
        self.assertEqual(obj_thresholds[1].metric.name, "branin_b")
        self.assertEqual(obj_thresholds[0].op, ComparisonOp.LEQ)
        self.assertEqual(obj_thresholds[1].op, ComparisonOp.LEQ)
        self.assertFalse(obj_thresholds[0].relative)
        self.assertFalse(obj_thresholds[1].relative)

        # test with MBM
        exp = get_branin_experiment_with_multi_objective(
            has_optimization_config=True,
            with_batch=True,
            with_status_quo=True,
        )
        for trial in exp.trials.values():
            trial.mark_running(no_runner_required=True).mark_completed()
        exp.attach_data(
            get_branin_data_multi_objective(trial_indices=exp.trials.keys())
        )
        data = exp.fetch_data()
        adapter = TorchAdapter(
            search_space=exp.search_space,
            generator=BoTorchGenerator(),
            optimization_config=exp.optimization_config,
            transforms=Cont_X_trans + Y_trans,
            torch_device=torch.device("cuda" if cuda else "cpu"),
            experiment=exp,
            data=data,
            # [T143911996] The trials get ignored without fit_out_of_design.
            fit_out_of_design=True,
        )
        obj_thresholds = adapter.infer_objective_thresholds(
            search_space=exp.search_space,
            optimization_config=exp.optimization_config,
            fixed_features=None,
        )
        self.assertEqual(obj_thresholds[0].metric.name, "branin_a")
        self.assertEqual(obj_thresholds[1].metric.name, "branin_b")
        self.assertEqual(obj_thresholds[0].op, ComparisonOp.LEQ)
        self.assertEqual(obj_thresholds[1].op, ComparisonOp.LEQ)
        self.assertFalse(obj_thresholds[0].relative)
        self.assertFalse(obj_thresholds[1].relative)

    def test_best_point(self) -> None:
        exp = get_branin_experiment_with_multi_objective(
            has_optimization_config=True, with_batch=True
        )
        for trial in exp.trials.values():
            trial.mark_running(no_runner_required=True).mark_completed()
        exp.attach_data(
            get_branin_data_multi_objective(trial_indices=exp.trials.keys())
        )
        adapter = TorchAdapter(
            search_space=exp.search_space,
            generator=MultiObjectiveLegacyBoTorchGenerator(),
            optimization_config=exp.optimization_config,
            transforms=[],
            experiment=exp,
            data=exp.fetch_data(),
        )
        self.assertIsNone(adapter.model_best_point())



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest import mock

import numpy as np
from ax.adapter.base import Adapter
from ax.adapter.registry import Generators
from ax.core import Experiment
from ax.core.arm import Arm
from ax.core.metric import Metric
from ax.core.observation import Observation, ObservationData
from ax.core.search_space import SearchSpace
from ax.generators.base import Generator
from ax.metrics.branin import BraninMetric
from ax.plot.scatter import tile_fitted, tile_observations
from ax.runners.synthetic import SyntheticRunner
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_arms,
    get_branin_experiment,
    get_branin_search_space,
    get_data,
    get_experiment_with_data,
)


def get_adapter(with_status_quo: bool) -> Adapter:
    experiment = get_branin_experiment(
        with_completed_batch=True, with_status_quo=with_status_quo
    )
    adapter = Adapter(experiment=experiment, generator=Generator())
    adapter._predict = mock.MagicMock(
        "ax.adapter.base.Adapter._predict",
        autospec=True,
        return_value=[
            ObservationData(
                metric_names=["branin"],
                means=np.array([1.0]),
                covariance=np.array([[1.0]]),
            )
        ],
    )
    return adapter


class TileFittedTest(TestCase):
    def test_TileFitted(self) -> None:
        adapter = get_adapter(with_status_quo=False)

        # Should throw if `status_quo_arm` is None and rel=True
        with self.assertRaises(ValueError):
            tile_fitted(adapter, rel=True)
        tile_fitted(adapter, rel=False)

        adapter = get_adapter(with_status_quo=True)
        config = tile_fitted(adapter, rel=True)
        self.assertIsNotNone(config)

        for key in ["layout", "data"]:
            self.assertIn(key, config.data)

        # Layout
        for key in [
            "annotations",
            "margin",
            "hovermode",
            "updatemenus",
            "font",
            "width",
            "height",
            "legend",
        ]:
            self.assertIn(key, config.data["layout"])

        # Data
        data = config.data["data"][0]
        arm_names = [arm.name for arm in adapter._experiment.trials[0].arms]
        # Bring SQ to the front.
        arm_names.insert(0, arm_names.pop(arm_names.index("status_quo")))
        self.assertEqual(data["x"], arm_names)
        self.assertEqual(data["y"], [0.0] * len(arm_names))
        self.assertEqual(data["type"], "scatter")
        for text, arm in zip(data["text"], arm_names):
            self.assertIn(f"Arm {arm}", text)
            self.assertIn("[-277.186%, 277.186%]", text)
            self.assertIn("0.0%", text)

        for key in [
            "type",
            "x",
            "y",
            "marker",
            "mode",
            "name",
            "text",
            "hoverinfo",
            "error_y",
            "visible",
            "legendgroup",
            "showlegend",
            "xaxis",
            "yaxis",
        ]:
            self.assertIn(key, data)

    def test_TileFittedOutOfDesignDataSelector(self) -> None:
        exp = Experiment(
            name="test3",
            search_space=get_branin_search_space(),
            tracking_metrics=[BraninMetric(name="b", param_names=["x1", "x2"])],
            runner=SyntheticRunner(),
        )
        batch = exp.new_batch_trial()
        arms = get_branin_arms(n=4, seed=0)
        arms.append(Arm(name="status_quo", parameters={"x1": None, "x2": None}))
        batch.add_arms_and_weights(arms=arms)
        batch.run()
        # batch.mark_running()
        batch.mark_completed()
        batch = exp.new_batch_trial()
        arms = get_branin_arms(n=4, seed=1)
        arms.append(Arm(name="status_quo", parameters={"x1": None, "x2": None}))
        batch.add_arms_and_weights(arms=arms)
        batch.run()
        batch.mark_completed()
        data0 = get_data(
            metric_name="b",
            trial_index=0,
            num_non_sq_arms=4,
        )
        data1 = get_data(
            metric_name="b",
            trial_index=1,
            num_non_sq_arms=4,
        )
        exp.attach_data(data0)
        exp.attach_data(data1)

        adapter = Generators.THOMPSON(experiment=exp, data=exp.lookup_data())

        def data_selector(obs: Observation) -> bool:
            return obs.features.trial_index != 0 or obs.arm_name == "status_quo"

        tile_fitted(adapter, rel=False)
        tile_fitted(
            adapter,
            rel=False,
            data_selector=data_selector,
        )


class TileObservationsTest(TestCase):
    def test_TileObservations(self) -> None:
        exp = get_experiment_with_data()
        exp.trials[0].run()
        exp.trials[0].mark_completed()
        exp.add_tracking_metric(Metric("ax_test_metric"))
        exp.search_space = SearchSpace(
            parameters=list(exp.search_space.parameters.values())
        )
        config = tile_observations(experiment=exp, arm_names=["0_0", "0_1"], rel=False)

        for key in ["layout", "data"]:
            self.assertIn(key, config.data)

        # Layout
        for key in [
            "annotations",
            "margin",
            "hovermode",
            "updatemenus",
            "font",
            "width",
            "height",
            "legend",
        ]:
            self.assertIn(key, config.data["layout"])

        self.assertEqual(
            config.data["layout"]["annotations"][0]["text"], "ax_test_metric"
        )

        # Data
        self.assertEqual(config.data["data"][0]["x"], ["0_0", "0_1"])
        self.assertEqual(config.data["data"][0]["y"], [3.0, 2.0])
        self.assertEqual(config.data["data"][0]["type"], "scatter")
        self.assertIn("Arm 0_0", config.data["data"][0]["text"][0])

        label_dict = {"ax_test_metric": "mapped_name"}
        config = tile_observations(
            experiment=exp, arm_names=["0_0", "0_1"], rel=False, label_dict=label_dict
        )
        self.assertEqual(config.data["layout"]["annotations"][0]["text"], "mapped_name")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import numpy as np
from ax.api.configs import ChoiceParameterConfig, RangeParameterConfig

from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    Parameter,
    ParameterType as CoreParameterType,
    RangeParameter,
)
from ax.exceptions.core import UserInputError


def parameter_from_config(
    config: RangeParameterConfig | ChoiceParameterConfig,
) -> Parameter:
    """
    Create a RangeParameter, ChoiceParameter, or FixedParameter from a ParameterConfig.
    """

    if isinstance(config, RangeParameterConfig):
        lower, upper = config.bounds

        # TODO[mpolson64] Add support for RangeParameterConfig.step_size native to
        # RangeParameter instead of converting to ChoiceParameter
        if (step_size := config.step_size) is not None:
            if not (config.scaling == "linear" or config.scaling is None):
                raise UserInputError(
                    "Non-linear parameter scaling is not supported when using "
                    "step_size."
                )

            if (upper - lower) % step_size != 0:
                raise UserInputError(
                    "The range of the parameter must be evenly divisible by the "
                    "step size."
                )

            return ChoiceParameter(
                name=config.name,
                parameter_type=_parameter_type_converter(config.parameter_type),
                values=[*np.arange(lower, upper + step_size, step_size)],
                is_ordered=True,
                sort_values=False,  # already sorted by np.arange.
            )

        return RangeParameter(
            name=config.name,
            parameter_type=_parameter_type_converter(config.parameter_type),
            lower=lower,
            upper=upper,
            log_scale=config.scaling == "log",
        )

    else:
        # If there is only one value, create a FixedParameter instead of a
        # ChoiceParameter
        if len(config.values) == 1:
            return FixedParameter(
                name=config.name,
                parameter_type=_parameter_type_converter(config.parameter_type),
                value=config.values[0],
                # pyre-fixme[6] Variance issue caused by FixedParameter.dependents
                # using List instead of immutable container type.
                dependents=config.dependent_parameters,
            )

        return ChoiceParameter(
            name=config.name,
            parameter_type=_parameter_type_converter(config.parameter_type),
            # pyre-fixme[6] Variance issue caused by ChoiceParameter.value using List
            # instead of immutable container type.
            values=config.values,
            is_ordered=config.is_ordered,
            # pyre-fixme[6] Variance issue caused by ChoiceParameter.dependents using
            # List instead of immutable container type.
            dependents=config.dependent_parameters,
            sort_values=config.parameter_type != "str",  # Matches default behavior.
        )


def _parameter_type_converter(parameter_type: str) -> CoreParameterType:
    """
    Convert from an API ParameterType to a core Ax ParameterType.
    """

    if parameter_type == "bool":
        return CoreParameterType.BOOL
    elif parameter_type == "float":
        return CoreParameterType.FLOAT
    elif parameter_type == "int":
        return CoreParameterType.INT
    elif parameter_type == "str":
        return CoreParameterType.STRING
    else:
        raise UserInputError(f"Unsupported parameter type {parameter_type}.")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Literal

from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import (
    AnalysisCard,
    AnalysisCardBase,
    AnalysisCardGroup,
    ErrorAnalysisCard,
)
from ax.analysis.plotly.sensitivity import SensitivityAnalysisPlot
from ax.analysis.plotly.surface.contour import (
    CONTOUR_CARDGROUP_SUBTITLE,
    CONTOUR_CARDGROUP_TITLE,
    ContourPlot,
)
from ax.analysis.plotly.surface.slice import (
    SLICE_CARDGROUP_SUBTITLE,
    SLICE_CARDGROUP_TITLE,
    SlicePlot,
)
from ax.analysis.plotly.utils import select_metric
from ax.core.experiment import Experiment
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import assert_is_instance, override

TS_CARDGROUP_TITLE = (
    "Top Surfaces Analysis: Parameter sensitivity, slice, and contour plots"
)

TS_CARDGROUP_SUBTITLE = (
    "The top surfaces analysis displays three analyses in one. First, it shows "
    "parameter sensitivities, which shows the sensitivity of the metrics in the "
    "experiment to the most important parameters. Subsetting to only the most "
    "important parameters, it then shows slice plots and contour plots for each "
    "metric in the experiment, displaying the relationship between the metric and "
    "the most important parameters. "
)


class TopSurfacesAnalysis(Analysis):
    def __init__(
        self,
        metric_name: str | None = None,
        order: Literal["first", "second", "total"] = "second",
        top_k: int = 5,
        relativize: bool = False,
    ) -> None:
        self.metric_name = metric_name
        self.order = order
        self.top_k = top_k
        self.relativize = relativize

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardGroup:
        if experiment is None:
            raise UserInputError(
                "TopSurfacesAnalysis requires an Experiment to compute."
            )

        if self.metric_name is not None:
            metric_name = self.metric_name
        else:
            metric_name = select_metric(experiment=experiment)

        # Process the sensitivity analysis card to find the top K surfaces which
        # consist exclusively of tunable parameters (i.e. no fixed parameters, task
        # parameters, or OneHot parameters).
        sensitivity_analysis_card = assert_is_instance(
            SensitivityAnalysisPlot(
                metric_names=[metric_name],
                order=self.order,
                top_k=self.top_k,
            ).compute(
                experiment=experiment,
                generation_strategy=generation_strategy,
                adapter=adapter,
            ),
            AnalysisCard,
        )
        children: list[AnalysisCardBase] = [sensitivity_analysis_card]

        sensitivity_df = sensitivity_analysis_card.df.copy()
        filtered_df = sensitivity_df[
            sensitivity_df["parameter_name"].apply(
                lambda x: all(
                    name in experiment.search_space.tunable_parameters
                    for name in x.split(" & ")
                )
            )
        ]
        sorted_df = filtered_df.sort_values(by="sensitivity", key=abs, ascending=False)
        top_k = sorted_df.head(self.top_k)
        top_surfaces = top_k["parameter_name"].to_list()

        surface_cards_or_error = [
            _compute_surface_plot(
                surface_name=surface_name,
                metric_name=metric_name,
                experiment=experiment,
                generation_strategy=generation_strategy,
                adapter=adapter,
                relativize=self.relativize,
            )
            for surface_name in top_surfaces
        ]

        # Filter out any ErrorAnalysisCards (i.e. failed to compute). When these
        # occur their presence will be logged by compute_or_error_card in
        # _compute_surface_plot so it's safe to filter them here
        surface_cards = [
            card
            for card in surface_cards_or_error
            if not isinstance(card, ErrorAnalysisCard)
        ]

        slice_cards = [
            assert_is_instance(card, AnalysisCard)
            for card in surface_cards
            if card.name == "SlicePlot"
        ]

        if len(slice_cards) > 0:
            children.append(
                AnalysisCardGroup(
                    name="TopSurfaceAnalysisSlicePlots",
                    title=SLICE_CARDGROUP_TITLE,
                    subtitle=SLICE_CARDGROUP_SUBTITLE,
                    children=slice_cards,
                )
            )

        contour_cards = [
            assert_is_instance(card, AnalysisCard)
            for card in surface_cards
            if card.name == "ContourPlot"
        ]

        if len(contour_cards) > 0:
            children.append(
                AnalysisCardGroup(
                    name="TopSurfaceAnalysisContourPlots",
                    title=CONTOUR_CARDGROUP_TITLE,
                    subtitle=CONTOUR_CARDGROUP_SUBTITLE,
                    children=contour_cards,
                )
            )

        return self._create_analysis_card_group(
            title=TS_CARDGROUP_TITLE,
            subtitle=TS_CARDGROUP_SUBTITLE,
            children=children,
        )


def _compute_surface_plot(
    surface_name: str,
    metric_name: str,
    experiment: Experiment | None,
    generation_strategy: GenerationStrategy | None,
    adapter: Adapter | None,
    relativize: bool = False,
) -> AnalysisCardBase:
    """Computes either a Slice or Contour plot for a given surface.
    Args:
        surface_name: The name of the parameter to plot. Either a single parameter or
            two parameters delimited by " & ". This determines whether a Slice or
            Contour plot is computed.
        experiment: The experiment to plot.
        generation_strategy: The generation strategy to plot.
        adapter: The adapter to plot.

    Returns:
        A PlotlyAnalysisCard containing the plot.
    """

    if "&" in surface_name:
        x_parameter_name, y_parameter_name = surface_name.split(" & ")
        analysis = ContourPlot(
            x_parameter_name=x_parameter_name,
            y_parameter_name=y_parameter_name,
            metric_name=metric_name,
            relativize=relativize,
        )

    else:
        analysis = SlicePlot(
            parameter_name=surface_name, metric_name=metric_name, relativize=relativize
        )

    return analysis.compute_or_error_card(
        experiment=experiment,
        generation_strategy=generation_strategy,
        adapter=adapter,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.analysis.plotly.surface.slice import compute_slice_adhoc, SlicePlot
from ax.core.trial import Trial
from ax.exceptions.core import UserInputError
from ax.service.ax_client import AxClient, ObjectiveProperties
from ax.utils.common.testutils import TestCase
from ax.utils.testing.mock import mock_botorch_optimize

from pyre_extensions import assert_is_instance, none_throws


class TestSlicePlot(TestCase):
    @mock_botorch_optimize
    def setUp(self) -> None:
        super().setUp()
        self.client = AxClient()
        self.client.create_experiment(
            is_test=True,
            name="foo",
            parameters=[
                {
                    "name": "x",
                    "type": "range",
                    "bounds": [-1.0, 1.0],
                }
            ],
            objectives={"bar": ObjectiveProperties(minimize=True)},
        )

        for _ in range(10):
            parameterization, trial_index = self.client.get_next_trial()
            self.client.complete_trial(
                trial_index=trial_index, raw_data={"bar": parameterization["x"] ** 2}
            )

    def test_compute(self) -> None:
        analysis = SlicePlot(parameter_name="x", metric_name="bar")

        # Test that it fails if no Experiment is provided
        with self.assertRaisesRegex(UserInputError, "requires an Experiment"):
            analysis.compute()
        # Test that it fails if no GenerationStrategy is provided
        with self.assertRaisesRegex(
            UserInputError, "Must provide either a GenerationStrategy or an Adapter"
        ):
            analysis.compute(experiment=self.client.experiment)

        card = analysis.compute(
            experiment=self.client.experiment,
            generation_strategy=self.client.generation_strategy,
        )
        self.assertEqual(
            card.name,
            "SlicePlot",
        )
        self.assertEqual(card.title, "x vs. bar")
        self.assertEqual(
            card.subtitle,
            (
                "The slice plot provides a one-dimensional view of predicted "
                "outcomes for bar as a function of a single parameter, "
                "while keeping all other parameters fixed at their status_quo "
                "value (or mean value if status_quo is unavailable). "
                "This visualization helps in understanding the sensitivity and "
                "impact of changes in the selected parameter on the predicted "
                "metric outcomes."
            ),
        )
        self.assertEqual(
            {*card.df.columns},
            {"x", "bar_mean", "bar_sem", "sampled", "trial_index", "arm_name"},
        )
        self.assertIsNotNone(card.blob)

        # Assert that any row where sampled is True has a value of x that is
        # sampled in at least one trial.
        x_values_sampled = {
            none_throws(assert_is_instance(trial, Trial).arm).parameters["x"]
            for trial in self.client.experiment.trials.values()
        }
        self.assertTrue(
            card.df.apply(
                lambda row: row["x"] in x_values_sampled if row["sampled"] else True,
                axis=1,
            ).all()
        )

    def test_compute_adhoc(self) -> None:
        (card,) = compute_slice_adhoc(
            parameter_name="x",
            metric_name="bar",
            experiment=self.client.experiment,
            generation_strategy=self.client.generation_strategy,
        ).flatten()
        self.assertEqual(
            card.name,
            "SlicePlot",
        )
        self.assertEqual(card.title, "x vs. bar")
        self.assertEqual(
            card.subtitle,
            (
                "The slice plot provides a one-dimensional view of predicted "
                "outcomes for bar as a function of a single parameter, "
                "while keeping all other parameters fixed at their status_quo "
                "value (or mean value if status_quo is unavailable). "
                "This visualization helps in understanding the sensitivity and "
                "impact of changes in the selected parameter on the predicted "
                "metric outcomes."
            ),
        )
        self.assertEqual(
            {*card.df.columns},
            {"x", "bar_mean", "bar_sem", "sampled", "arm_name", "trial_index"},
        )
        self.assertIsNotNone(card.blob)

        # Assert that any row where sampled is True has a value of x that is
        # sampled in at least one trial.
        x_values_sampled = {
            none_throws(assert_is_instance(trial, Trial).arm).parameters["x"]
            for trial in self.client.experiment.trials.values()
        }
        self.assertTrue(
            card.df.apply(
                lambda row: row["x"] in x_values_sampled if row["sampled"] else True,
                axis=1,
            ).all()
        )

    def test_trial_status_filtering(self) -> None:
        trial_index = self.client.experiment.new_trial().index
        self.client.experiment.trials[trial_index].mark_abandoned()

        analysis = SlicePlot(parameter_name="x", metric_name="bar")

        card = analysis.compute(
            experiment=self.client.experiment,
            generation_strategy=self.client.generation_strategy,
        )
        self.assertNotIn(
            trial_index,
            card.df["trial_index"].values,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Mapping, Sequence

import numpy as np

import pandas as pd
from ax.adapter.base import Adapter

from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCardBase
from ax.analysis.plotly.color_constants import (
    BOTORCH_COLOR_SCALE,
    CONSTRAINT_VIOLATION_RED,
)
from ax.analysis.plotly.plotly_analysis import create_plotly_analysis_card
from ax.analysis.plotly.utils import (
    BEST_LINE_SETTINGS,
    get_arm_tooltip,
    get_trial_trace_name,
    LEGEND_BASE_OFFSET,
    LEGEND_POSITION,
    MARGIN_REDUCUTION,
    MULTIPLE_CANDIDATE_TRIALS_LEGEND,
    SINGLE_CANDIDATE_TRIAL_LEGEND,
    trial_index_to_color,
    truncate_label,
    X_TICKER_SCALING_FACTOR,
    Z_SCORE_95_CI,
)
from ax.analysis.utils import (
    extract_relevant_adapter,
    get_lower_is_better,
    POSSIBLE_CONSTRAINT_VIOLATION_THRESHOLD,
    prepare_arm_data,
    update_metric_names_if_using_p_feasible,
)
from ax.core.arm import Arm
from ax.core.base_trial import sort_by_trial_index_and_arm_name
from ax.core.experiment import Experiment
from ax.core.trial_status import FAILED_ABANDONED_CANDIDATE_STATUSES, TrialStatus
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from plotly import graph_objects as go
from pyre_extensions import override

CARDGROUP_TITLE = "Metric Effects: Values of key metrics for all arms in the experiment"

PREDICTED_EFFECTS_CARDGROUP_SUBTITLE = (
    "These plots visualize predictions of the 'true' metric changes for each arm, "
    "based on Ax's model. Since Ax applies Empirical Bayes shrinkage to adjust for "
    "noise and also accounts for non-stationarity in the data, predicted metric "
    "effects will not match raw observed data perfectly, but will be more "
    "representative of the reproducible effects that will manifest in a long-term "
    "validation experiment. <br><br>"
    "NOTE: Flat predictions across arms indicate that the model predicts that "
    "none of the arms had a sufficient effect on the metric, meaning that if you "
    "re-ran the experiment, the delta you would see would be small and fall "
    "within the confidence interval indicated in the plot. In other words, this "
    "indicates that according to the model, the raw observed effects on this metric "
    "are primarily noise."
)

RAW_EFFECTS_CARDGROUP_SUBTITLE = (
    "These plots visualize the raw data on the effects we observed from "
    "previously-run arms on a specific metric, providing insights into "
    "their performance. These plots allow one to compare and contrast the "
    "effectiveness of different arms, highlighting which configurations have yielded "
    "the most favorable outcomes."
)

# If there are more than this many non-candidate arms to be plotted, change the way we
# organize the xaxis and its labels to improve legibility.
ARM_LABEL_THRESHOLD = 20


class ArmEffectsPlot(Analysis):
    """
    Plot the effects of each arm in an experiment on a given metric. Effects may be
    either the raw observed effects, or the predicted effects using a model. The
    latter is often more trustworthy (and leads to better reproducibility) than using
    the raw data, especially when model fit is good and in high-noise settings.

    Each arm is represented by a point on the plot with 95% confidence intervals. The
    color of the point indicates the status of the arm's trial (e.g. RUNNING, SUCCEDED,
    FAILED). Arms which are likely to violate a constraint (i.e. according to either
    the raw or modeled effects, the probability all constraints are satisfied is < 5%)
    are marked with a red outline. Each arm also has a hover tooltip with additional
    information.

    The DataFrame computed will contain one row per arm and the following columns:
        - trial_index: The trial index of the arm
        - trial_status: The status of the trial
        - arm_name: The name of the arm
        - generation_node: The name of the ``GenerationNode`` that generated the arm
        - p_feasible: The probability that the arm is feasible (does not violate any
            constraints)
        - **METRIC_NAME_mean: The observed mean of the metric specified
        - **METRIC_NAME_sem: The observed sem of the metric specified
    """

    def __init__(
        self,
        metric_names: Sequence[str] | None = None,
        use_model_predictions: bool = True,
        relativize: bool = False,
        trial_index: int | None = None,
        trial_statuses: Sequence[TrialStatus] | None = None,
        additional_arms: Sequence[Arm] | None = None,
        labels: Mapping[str, str] | None = None,
        show_cumulative_best: bool = False,
    ) -> None:
        """
        Args:
            metric_names: The names of the metrics to include in the plot. If not
                specified, all metrics in the experiment will be used.
            use_model_predictions: Whether to use model predictions or observed data.
                If ``True``, the plot will show the predicted effects of each arm based
                on the model. If ``False``, the plot will show the observed effects of
                each arm. The latter is often less trustworthy than the former,
                especially when model fit is good and in high-noise settings.
            relativize: Whether to relativize the effects of each arm against the status
                quo arm. If multiple status quo arms are present, relativize each arm
                against the status quo arm from the same trial.
            trial_index: If present, only use arms from the trial with the given index.
            additional_arms: If present, include these arms in the plot in addition to
                the arms in the experiment. These arms will be marked as belonging to a
                trial with index -1.
            labels: A mapping from metric names to labels to use in the plot. If a label
                is not provided for a metric, the metric name will be used.
            show_cumulative_best: Whether to draw a line through the best point seen so
                far during the optimization.
        """

        self.metric_names = metric_names
        self.use_model_predictions = use_model_predictions
        self.relativize = relativize
        self.trial_index = trial_index
        self.trial_statuses = trial_statuses
        self.additional_arms = additional_arms
        self.labels: Mapping[str, str] = labels or {}
        self.show_cumulative_best = show_cumulative_best

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardBase:
        if experiment is None:
            raise UserInputError("ArmEffectsPlot requires an Experiment.")

        metric_names = self.metric_names or [*experiment.metrics.keys()]

        if self.use_model_predictions:
            relevant_adapter = extract_relevant_adapter(
                experiment=experiment,
                generation_strategy=generation_strategy,
                adapter=adapter,
            )
        else:
            relevant_adapter = None
        metric_names_to_plot = metric_names
        # if using p_feasible, we need to get the data for the metrics involved
        # in constraints even though we don't plot them
        metric_names = update_metric_names_if_using_p_feasible(
            metric_names=metric_names_to_plot, experiment=experiment
        )
        df = prepare_arm_data(
            experiment=experiment,
            metric_names=metric_names,
            use_model_predictions=self.use_model_predictions,
            adapter=relevant_adapter,
            trial_index=self.trial_index,
            trial_statuses=self.trial_statuses,
            additional_arms=self.additional_arms,
            relativize=self.relativize,
        )

        # Retrieve the metric labels from the mapping provided by the user, defaulting
        # to the metric name if no label is provided, truncated.
        metric_labels = {
            metric_name: self.labels.get(metric_name, truncate_label(label=metric_name))
            for metric_name in metric_names
        }

        cards = [
            create_plotly_analysis_card(
                name=self.__class__.__name__,
                title=(
                    f"{'Modeled' if self.use_model_predictions else 'Observed'} "
                    f"{'Relativized ' if self.relativize else ''}Arm "
                    f"Effects on {metric_labels[metric_name]}"
                    + (
                        f" for trial {self.trial_index}"
                        if self.trial_index is not None
                        else ""
                    )
                ),
                subtitle=_get_subtitle(
                    metric_label=metric_labels[metric_name],
                    use_model_predictions=self.use_model_predictions,
                    trial_index=self.trial_index,
                ),
                df=df[
                    [
                        "trial_index",
                        "trial_status",
                        "arm_name",
                        "generation_node",
                        "p_feasible_mean",
                        "p_feasible_sem",
                    ]
                    + (
                        [
                            f"{metric_name}_mean",
                            f"{metric_name}_sem",
                        ]
                        if metric_name != "p_feasible"
                        else []
                    )
                ].copy(),
                fig=_prepare_figure(
                    df=df,
                    metric_name=metric_name,
                    is_relative=self.relativize,
                    status_quo_arm_name=experiment.status_quo.name
                    if experiment.status_quo
                    else None,
                    metric_label=metric_labels[metric_name],
                    show_cumulative_best=self.show_cumulative_best,
                    lower_is_better=get_lower_is_better(
                        experiment=experiment, metric_name=metric_name
                    )
                    or False,
                ),
            )
            for metric_name in metric_names_to_plot
        ]

        return self._create_analysis_card_group_or_card(
            title=CARDGROUP_TITLE,
            subtitle=PREDICTED_EFFECTS_CARDGROUP_SUBTITLE
            if self.use_model_predictions
            else RAW_EFFECTS_CARDGROUP_SUBTITLE,
            children=cards,
        )


def compute_arm_effects_adhoc(
    experiment: Experiment,
    generation_strategy: GenerationStrategy | None = None,
    adapter: Adapter | None = None,
    metric_names: Sequence[str] | None = None,
    use_model_predictions: bool = True,
    relativize: bool = False,
    trial_index: int | None = None,
    trial_statuses: Sequence[TrialStatus] | None = None,
    additional_arms: Sequence[Arm] | None = None,
    labels: Mapping[str, str] | None = None,
    show_cumulative_best: bool = False,
) -> AnalysisCardBase:
    """
    Compute ArmEffectsPlot cards for the given experiment and either Adapter or
    GenerationStrategy.

    Note that cards are not saved to the database when computed adhoc -- they are only
    saved when computed as part of a call to ``Client.compute_analyses`` or equivalent.

    Args:
        experiment: The experiment to extract data from.
        metric_names: The names of the metrics to include in the plot. If not
            specified, all metrics in the experiment will be used.
        generation_strategy: The GenerationStrategy to use for predictions if
            use_model_predictions=True.
        adapter: The adapter to use for predictions if use_model_predictions=True.
        use_model_predictions: Whether to use model predictions or observed data.
            If ``True``, the plot will show the predicted effects of each arm based
            on the model. If ``False``, the plot will show the observed effects of
            each arm. The latter is often less trustworthy than the former,
            especially when model fit is good and in high-noise settings.
        relativize: Whether to relativize the effects of each arm against the status
            quo arm. If multiple status quo arms are present, relativize each arm
            against the status quo arm from the same trial.
        trial_index: If present, only use arms from the trial with the given index.
        additional_arms: If present, include these arms in the plot in addition to
            the arms in the experiment. These arms will be marked as belonging to a
            trial with index -1.
        labels: A mapping from metric names to labels to use in the plot. If a label
            is not provided for a metric, the metric name will be used.
        show_cumulative_best: Whether to draw a line through the best point seen so
            far during the optimization.
    """

    analysis = ArmEffectsPlot(
        metric_names=metric_names,
        use_model_predictions=use_model_predictions,
        relativize=relativize,
        trial_index=trial_index,
        trial_statuses=trial_statuses,
        additional_arms=additional_arms,
        labels=labels,
        show_cumulative_best=show_cumulative_best,
    )

    return analysis.compute(
        experiment=experiment,
        generation_strategy=generation_strategy,
        adapter=adapter,
    )


def _prepare_figure(
    df: pd.DataFrame,
    metric_name: str,
    is_relative: bool,
    status_quo_arm_name: str | None,
    metric_label: str,
    show_cumulative_best: bool,
    lower_is_better: bool,
) -> go.Figure:
    # Prepare separate scatter traces for each trial index. Each trace has (x, y)
    # points for the arms which we have a mean for in the provided dataframe.
    candidate_trial = df[df["trial_status"] == TrialStatus.CANDIDATE.name][
        "trial_index"
    ].max()
    # Filter out undesired trials like FAILED and ABANDONED trials from plot.
    trials = df[
        ~df["trial_status"].isin(
            [ts.name for ts in FAILED_ABANDONED_CANDIDATE_STATUSES]
        )
    ]["trial_index"].unique()

    # Check if candidate_trial is NaN and handle it
    trial_indices = list(trials)
    if not np.isnan(candidate_trial):
        trial_indices.append(candidate_trial)

    trials_list = trials.tolist()

    scatters = []
    scatter_trial_indices = []  # Track trial indices for each scatter
    # Sort the dataframe by trial index and arm name.
    # Non-default arm names (not digit + underscore + digit) are sorted to the front.
    # default arm names (e.g. 0_0, 1_8) are sorted that '0_1' < '0_5' < '0_10'
    df = sort_by_trial_index_and_arm_name(df=df)
    # Add a column combining trial_index and arm_name to be used as x mark
    df["x_key_order"] = df["trial_index"].astype(str) + ":" + df["arm_name"]
    arm_order = []
    arm_label = []

    # Track trials that get included in the plot
    num_candidate_trials = 0
    candidate_trial_marker = None

    for trial_index in trial_indices:
        trial_df = df[df["trial_index"] == trial_index]
        xy_df = trial_df[~trial_df[f"{metric_name}_mean"].isna()]
        # Skip trials with no valid data points as they will not end up in the plot
        if xy_df.empty:
            continue

        arm_order = arm_order + xy_df["x_key_order"].to_list()
        arm_label = arm_label + xy_df["arm_name"].to_list()
        if not trial_df[f"{metric_name}_sem"].isna().all():
            error_y = {
                "type": "data",
                "array": Z_SCORE_95_CI * xy_df[f"{metric_name}_sem"],
                "color": trial_index_to_color(
                    trial_df=trial_df,
                    trials_list=trials_list,
                    trial_index=trial_index,
                    transparent=True,
                ),
            }
        else:
            error_y = None

        marker = {
            "color": trial_index_to_color(
                trial_df=trial_df,
                trials_list=trials_list,
                trial_index=trial_index,
                transparent=False,
            ),
            "line": {
                "width": trial_df.apply(
                    lambda row: 2
                    if row["p_feasible_mean"] < POSSIBLE_CONSTRAINT_VIOLATION_THRESHOLD
                    else 0,
                    axis=1,
                ),
                "color": CONSTRAINT_VIOLATION_RED,
            },
        }

        if trial_df["trial_status"].iloc[0] == TrialStatus.CANDIDATE.name:
            num_candidate_trials += 1
            candidate_trial_marker = marker

        text = trial_df.apply(
            lambda row: get_arm_tooltip(row=row, metric_names=[metric_name]), axis=1
        )

        scatters.append(
            go.Scatter(
                x=xy_df["x_key_order"],
                y=xy_df[f"{metric_name}_mean"],
                error_y=error_y,
                mode="markers",
                marker=marker,
                name=get_trial_trace_name(trial_index=trial_index),
                showlegend=False,  # Will be set after determining use_colorscale
                hoverinfo="text",
                text=text,
                legendgroup="candidate_trials"
                if trial_df["trial_status"].iloc[0] == TrialStatus.CANDIDATE.name
                else None,
            )
        )
        scatter_trial_indices.append(trial_index)

    # Determine use_colorscale based on the number of included arms
    use_colorscale = len(arm_label) > ARM_LABEL_THRESHOLD

    # Update markers and legend settings based on use_colorscale
    for scatter, trial_index in zip(scatters, scatter_trial_indices):
        trial_df = df[df["trial_index"] == trial_index]

        if use_colorscale:
            # Add colorscale settings to marker
            scatter.marker.update(
                {
                    "colorscale": BOTORCH_COLOR_SCALE,
                    "showscale": True,
                    "cmin": min(scatter_trial_indices),
                    "cmax": max(scatter_trial_indices),
                    "colorbar": {
                        "title": "Trial Index",
                        "orientation": "h",
                        "x": 0.4,
                        "xanchor": "center",
                        "y": -0.30,
                        "yanchor": "top",
                    },
                }
            )
        else:
            # Show legend for all non-candidate trials when not using colorscale
            scatter.showlegend = (
                trial_df["trial_status"].iloc[0] != TrialStatus.CANDIDATE.name
            )

    # get the max length of x-ticker (arm name) to set the xaxis label and
    # legend position
    # This assumes the x-tickers are rotated 90 degrees (vertical) so legend
    # will be always below the x-label ('Arm Name').
    max_label_len = max(len(label) for label in arm_label)
    legend_y = LEGEND_BASE_OFFSET - (max_label_len / X_TICKER_SCALING_FACTOR)

    legend_position = LEGEND_POSITION.copy()
    if use_colorscale:
        # Position candidate legend to the right of the colorscale
        legend_position.update(
            {
                "orientation": "v",
                "yanchor": "top",
                "y": -0.33,
                "xanchor": "left",
                "x": 0.9,
            }
        )
    else:
        legend_position["y"] = legend_y

    figure = go.Figure(data=scatters)
    figure.update_layout(
        xaxis_title="Arm Name",
        yaxis_title=metric_label,
        yaxis_tickformat=".2%" if is_relative else None,
        legend=legend_position,
        margin=MARGIN_REDUCUTION,
        xaxis={
            "tickvals": [
                arm_order[i * (len(arm_order) // ARM_LABEL_THRESHOLD)]
                for i in range(ARM_LABEL_THRESHOLD)
            ]
            if use_colorscale
            else arm_order,
            "ticktext": [
                arm_label[i * (len(arm_label) // ARM_LABEL_THRESHOLD)]
                for i in range(ARM_LABEL_THRESHOLD)
            ]
            if use_colorscale
            else arm_label,
        },
    )

    # Add candidate trial legend at the end
    if num_candidate_trials > 0:
        figure.add_trace(
            go.Scatter(
                x=[None],
                y=[None],
                mode="markers",
                marker=candidate_trial_marker,
                name=SINGLE_CANDIDATE_TRIAL_LEGEND
                if num_candidate_trials == 1
                else MULTIPLE_CANDIDATE_TRIALS_LEGEND,
                showlegend=True,
                hoverinfo="skip",
                legendgroup="candidate_trials",
            )
        )

    # Add a horizontal line for the status quo.
    if status_quo_arm_name in df["arm_name"].values:
        # In relativized plots the status quo is always 0% on the y-axis.
        if is_relative:
            y = 0
        else:
            y = df[df["arm_name"] == status_quo_arm_name][f"{metric_name}_mean"].iloc[0]

        figure.add_shape(
            type="line",
            xref="paper",
            x0=0,
            y0=y,
            x1=1,
            y1=y,
            line={"color": "gray", "dash": "dot"},
        )

    # Add a red circle with no fill if any arms are marked as possibly infeasible.
    if (df["p_feasible_mean"] < POSSIBLE_CONSTRAINT_VIOLATION_THRESHOLD).any():
        legend_trace = go.Scatter(
            # None here allows us to place a legend item without corresponding points
            x=[None],
            y=[None],
            mode="markers",
            marker={
                "color": "rgba(0, 0, 0, 0)",
                "line": {"width": 2, "color": "red"},
            },
            name="Possible Constraint Violation",
        )

        figure.add_trace(legend_trace)

    if show_cumulative_best:
        # Sort and filter arms which are eligable to the best point (i.e. that are not
        # likely to violate constraints).
        zipped = [
            *zip(
                *[
                    (
                        arm_name,
                        df[df["arm_name"] == arm_name][f"{metric_name}_mean"].iloc[0],
                    )
                    for arm_name in arm_label
                    if df[df["arm_name"] == arm_name]["p_feasible_mean"].iloc[0]
                    >= POSSIBLE_CONSTRAINT_VIOLATION_THRESHOLD
                ]
            )
        ]

        # If there are no arms which are not likely to violate constraints, return the
        # figure as is, without adding a best point line.
        if len(zipped) != 2:
            return figure

        sorted_arms, sorted_y = zipped

        best_y = (
            np.minimum.accumulate(sorted_y)
            if lower_is_better
            else np.maximum.accumulate(sorted_y)
        )

        # Add a line for the best point seen so far.
        figure.add_trace(go.Scatter(x=sorted_arms, y=best_y, **BEST_LINE_SETTINGS))

    return figure


def _get_subtitle(
    metric_label: str,
    use_model_predictions: bool,
    trial_index: int | None = None,
) -> str:
    first_clause = (
        f"{'Modeled' if use_model_predictions else 'Observed'} effects on "
        f"{metric_label}"
    )
    trial_clause = f" for Trial {trial_index}." if trial_index is not None else ""
    first_sentence = f"{first_clause}{trial_clause}."

    if use_model_predictions:
        return (
            f"{first_sentence} This plot visualizes predictions of the "
            "true metric changes for each arm based on Ax's model. This is the "
            "expected delta you would expect if you (re-)ran that arm. This plot helps "
            "in anticipating the outcomes and performance of arms based on the model's "
            "predictions. Note, flat predictions across arms indicate that the model "
            "predicts that there is no effect, meaning if you were to re-run the "
            "experiment, the delta you would see would be small and fall within the "
            "confidence interval indicated in the plot."
        )

    return (
        f"{first_sentence} This plot visualizes the effects from previously-run arms "
        "on a specific metric, providing insights into their performance. This plot "
        "allows one to compare and contrast the effectiveness of different arms, "
        "highlighting which configurations have yielded the most favorable outcomes. "
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Module containing a registry of standard models (and generators, samplers etc.)
such as Sobol generator, GP+EI, Thompson sampler, etc.
"""

from __future__ import annotations

import warnings
from collections.abc import Mapping, Sequence
from enum import Enum
from inspect import isfunction, signature
from logging import Logger
from typing import Any, NamedTuple

from ax.adapter.base import Adapter, DataLoaderConfig
from ax.adapter.discrete import DiscreteAdapter
from ax.adapter.random import RandomAdapter
from ax.adapter.torch import TorchAdapter
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.bilog_y import BilogY
from ax.adapter.transforms.choice_encode import (
    ChoiceToNumericChoice,
    OrderedChoiceToIntegerRange,
)
from ax.adapter.transforms.derelativize import Derelativize
from ax.adapter.transforms.int_range_to_choice import IntRangeToChoice
from ax.adapter.transforms.int_to_float import IntToFloat, LogIntToFloat
from ax.adapter.transforms.log import Log
from ax.adapter.transforms.logit import Logit
from ax.adapter.transforms.map_key_to_float import MapKeyToFloat
from ax.adapter.transforms.merge_repeated_measurements import MergeRepeatedMeasurements
from ax.adapter.transforms.one_hot import OneHot
from ax.adapter.transforms.relativize import Relativize
from ax.adapter.transforms.remove_fixed import RemoveFixed
from ax.adapter.transforms.search_space_to_choice import SearchSpaceToChoice
from ax.adapter.transforms.standardize_y import StandardizeY
from ax.adapter.transforms.stratified_standardize_y import StratifiedStandardizeY
from ax.adapter.transforms.task_encode import TaskChoiceToIntTaskChoice
from ax.adapter.transforms.transform_to_new_sq import TransformToNewSQ
from ax.adapter.transforms.trial_as_task import TrialAsTask
from ax.adapter.transforms.unit_x import UnitX
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.search_space import SearchSpace
from ax.exceptions.core import UserInputError
from ax.generators.base import Generator
from ax.generators.discrete.eb_ashr import EBAshr
from ax.generators.discrete.eb_thompson import EmpiricalBayesThompsonSampler
from ax.generators.discrete.full_factorial import FullFactorialGenerator
from ax.generators.discrete.thompson import ThompsonSampler
from ax.generators.random.sobol import SobolGenerator
from ax.generators.random.uniform import UniformGenerator
from ax.generators.torch.botorch import LegacyBoTorchGenerator
from ax.generators.torch.botorch_modular.generator import (
    BoTorchGenerator as ModularBoTorchGenerator,
)
from ax.generators.torch.botorch_modular.surrogate import ModelConfig, SurrogateSpec
from ax.generators.torch.cbo_sac import SACBO
from ax.utils.common.kwargs import (
    consolidate_kwargs,
    get_function_argument_names,
    get_function_default_arguments,
)
from ax.utils.common.logger import get_logger
from ax.utils.common.serialization import callable_from_reference, callable_to_reference
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.fully_bayesian_multitask import SaasFullyBayesianMultiTaskGP
from pyre_extensions import assert_is_instance, none_throws

logger: Logger = get_logger(__name__)

# This set of transforms uses continuous relaxation to handle discrete parameters.
# All candidate generation is done in the continuous space, and the generated
# candidates are rounded to fit the original search space. This is can be
# suboptimal when there are discrete parameters with a small number of options.
Cont_X_trans: list[type[Transform]] = [
    RemoveFixed,
    OrderedChoiceToIntegerRange,
    OneHot,
    IntToFloat,
    Log,
    Logit,
    UnitX,
]

# This is a modification of Cont_X_trans that aims to avoid continuous relaxation
# where possible. It replaces IntToFloat with LogIntToFloat, which is only transforms
# log-scale integer parameters, which still use continuous relaxation. Other discrete
# transforms will remain discrete. When used with MBM, a Normalize input transform
# will be added to replace the UnitX transform. This setup facilitates the use of
# optimize_acqf_mixed_alternating, which is a more efficient acquisition function
# optimizer for mixed discrete/continuous problems.
MBM_X_trans_base: list[type[Transform]] = [
    RemoveFixed,
    OrderedChoiceToIntegerRange,
    OneHot,
    LogIntToFloat,
    Log,
    Logit,
]
MBM_X_trans: list[type[Transform]] = [MapKeyToFloat, *MBM_X_trans_base]


Discrete_X_trans: list[type[Transform]] = [IntRangeToChoice]

EB_ashr_trans: list[type[Transform]] = [
    Derelativize,  # necessary to support relative constraints
    # scales data from multiple trials since we currently don't filter to single
    # trial data
    TransformToNewSQ,
    # Ensure we pass unique arms to EBAshr. This assumes treatment effects are
    # stationarity, but also should help with estimating the task-task covariance.
    MergeRepeatedMeasurements,
    SearchSpaceToChoice,
]

# TODO: @mgarrard remove this once non-gs methods are reaped
rel_EB_ashr_trans: list[type[Transform]] = [
    Relativize,
    MergeRepeatedMeasurements,
    SearchSpaceToChoice,
]

# This is a modification of MBM_X_trans that replaces OneHot and
# OrderedChoiceToIntegerRange with ChoiceToNumericChoice.
# This retains unordered choice parameters as a single parameter
# and uses MixedSingleTaskGP with CategoricalKernel in MBM to support them.
Mixed_transforms: list[type[Transform]] = [
    RemoveFixed,
    ChoiceToNumericChoice,
    LogIntToFloat,
    Log,
    Logit,
]

Y_trans: list[type[Transform]] = [Derelativize, BilogY, StandardizeY]

# Expected `List[Type[Transform]]` for 2nd anonymous parameter to
# call `list.__add__` but got `List[Type[SearchSpaceToChoice]]`.
TS_trans: list[type[Transform]] = Y_trans + [SearchSpaceToChoice]

MTGP_Y_trans: list[type[Transform]] = [
    Derelativize,
    TrialAsTask,
    StratifiedStandardizeY,
    TaskChoiceToIntTaskChoice,
]

# Single-type MTGP transforms
ST_MTGP_trans: list[type[Transform]] = Cont_X_trans + MTGP_Y_trans

MBM_MTGP_trans: list[type[Transform]] = MBM_X_trans + MTGP_Y_trans


class ModelSetup(NamedTuple):
    """A model setup defines a coupled combination of a model, an adapter,
    standard set of transforms, and standard adapter keyword arguments.
    This coupled combination yields a given standard modeling strategy in Ax,
    such as BoTorch GP+EI, a Thompson sampler, or a Sobol quasirandom generator.
    """

    adapter_class: type[Adapter]
    model_class: type[Generator]
    transforms: Sequence[type[Transform]]
    default_model_kwargs: Mapping[str, Any] | None = None
    standard_bridge_kwargs: Mapping[str, Any] | None = None
    not_saved_model_kwargs: Sequence[str] | None = None


"""A mapping of string keys that indicate a model, to the corresponding
model setup, which defines which model, adapter, transforms, and
standard arguments a given model requires.
"""
MODEL_KEY_TO_MODEL_SETUP: dict[str, ModelSetup] = {
    "BoTorch": ModelSetup(
        adapter_class=TorchAdapter,
        model_class=ModularBoTorchGenerator,
        transforms=MBM_X_trans + Y_trans,
    ),
    "Legacy_GPEI": ModelSetup(
        adapter_class=TorchAdapter,
        standard_bridge_kwargs={
            "data_loader_config": DataLoaderConfig(
                fit_only_completed_map_metrics=True,
            ),
        },
        model_class=LegacyBoTorchGenerator,
        transforms=Cont_X_trans + Y_trans,
    ),
    "EB": ModelSetup(
        adapter_class=DiscreteAdapter,
        model_class=EmpiricalBayesThompsonSampler,
        transforms=TS_trans,
    ),
    "EB_Ashr": ModelSetup(
        adapter_class=DiscreteAdapter,
        model_class=EBAshr,
        transforms=EB_ashr_trans,
    ),
    "Factorial": ModelSetup(
        adapter_class=DiscreteAdapter,
        model_class=FullFactorialGenerator,
        transforms=Discrete_X_trans,
    ),
    "Thompson": ModelSetup(
        adapter_class=DiscreteAdapter,
        model_class=ThompsonSampler,
        transforms=TS_trans,
    ),
    "Sobol": ModelSetup(
        adapter_class=RandomAdapter,
        model_class=SobolGenerator,
        transforms=Cont_X_trans,
    ),
    "Uniform": ModelSetup(
        adapter_class=RandomAdapter,
        model_class=UniformGenerator,
        transforms=Cont_X_trans,
    ),
    "ST_MTGP": ModelSetup(
        adapter_class=TorchAdapter,
        model_class=ModularBoTorchGenerator,
        transforms=MBM_MTGP_trans,
    ),
    "BO_MIXED": ModelSetup(
        adapter_class=TorchAdapter,
        model_class=ModularBoTorchGenerator,
        transforms=Mixed_transforms + Y_trans,
    ),
    "SAASBO": ModelSetup(
        adapter_class=TorchAdapter,
        model_class=ModularBoTorchGenerator,
        transforms=MBM_X_trans + Y_trans,
        default_model_kwargs={
            "surrogate_spec": SurrogateSpec(
                model_configs=[
                    ModelConfig(
                        botorch_model_class=SaasFullyBayesianSingleTaskGP, name="SAASBO"
                    )
                ]
            )
        },
    ),
    "SAAS_MTGP": ModelSetup(
        adapter_class=TorchAdapter,
        model_class=ModularBoTorchGenerator,
        transforms=MBM_MTGP_trans,
        default_model_kwargs={
            "surrogate_spec": SurrogateSpec(
                model_configs=[
                    ModelConfig(
                        botorch_model_class=SaasFullyBayesianMultiTaskGP,
                        name="SAAS_MTGP",
                    )
                ]
            )
        },
    ),
    "Contextual_SACBO": ModelSetup(
        adapter_class=TorchAdapter,
        model_class=SACBO,
        transforms=Cont_X_trans + Y_trans,
    ),
}


class GeneratorRegistryBase(Enum):
    """Base enum that provides instrumentation of `__call__` on enum values,
    for enums that link their values to `ModelSetup`-s like `Generators`.
    """

    @property
    def model_key_to_model_setup(self) -> dict[str, ModelSetup]:
        return MODEL_KEY_TO_MODEL_SETUP

    @property
    def model_class(self) -> type[Generator]:
        """Type of `Model` used for the given model+adapter setup."""
        return self.model_key_to_model_setup[self.value].model_class

    @property
    def adapter_class(self) -> type[Adapter]:
        """Type of `Adapter` used for the given model+adapter setup."""
        return self.model_key_to_model_setup[self.value].adapter_class

    def __call__(
        self,
        search_space: SearchSpace | None = None,
        experiment: Experiment | None = None,
        data: Data | None = None,
        silently_filter_kwargs: bool = False,
        **kwargs: Any,
    ) -> Adapter:
        if self.value not in self.model_key_to_model_setup:
            raise UserInputError(f"Unknown model {self.value}")
        model_setup_info = self.model_key_to_model_setup[self.value]
        model_class = model_setup_info.model_class
        adapter_class = model_setup_info.adapter_class
        if experiment is None:
            # Some Adapters used to accept search_space as the only input.
            # Temporarily support it with a deprecation warning.
            if (
                issubclass(adapter_class, (RandomAdapter, DiscreteAdapter))
                and search_space is not None
            ):
                warnings.warn(
                    "Passing in a `search_space` to initialize a generator from a "
                    "registry is being deprecated. `experiment` is now a required "
                    "input for initializing `Adapters`. Please use `experiment` "
                    "when initializing generators going forward. "
                    "Support for `search_space` will be removed in Ax 0.7.0.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                # Construct a dummy experiment for temporary support.
                experiment = Experiment(search_space=search_space)
            else:
                raise UserInputError(
                    "`experiment` is required to initialize a model from registry."
                )
        search_space = search_space or none_throws(experiment).search_space

        if not silently_filter_kwargs:
            # Check correct kwargs are present
            callables = (model_class, adapter_class)
            kwargs_to_check = {
                "search_space": search_space,
                "experiment": experiment,
                "data": data,
                **kwargs,
            }
            checked_kwargs = set()
            for fn in callables:
                params = signature(fn).parameters
                for kw in params.keys():
                    if kw in kwargs_to_check:
                        if kw in checked_kwargs:
                            logger.debug(
                                f"`{callables}` have duplicate keyword argument: {kw}."
                            )
                        else:
                            checked_kwargs.add(kw)

            # Check if kwargs contains keywords not exist in any callables
            extra_keywords = [kw for kw in kwargs.keys() if kw not in checked_kwargs]
            if len(extra_keywords) != 0:
                raise ValueError(
                    f"Arguments {extra_keywords} are not expected by any of {callables}"
                )

        # Create model with consolidated arguments: defaults + passed in kwargs.
        model_kwargs = consolidate_kwargs(
            kwargs_iterable=[
                get_function_default_arguments(model_class),
                model_setup_info.default_model_kwargs,
                kwargs,
            ],
            keywords=get_function_argument_names(model_class),
        )
        generator = model_class(**model_kwargs)

        # Create `Adapter`: defaults + standard kwargs + passed in kwargs.
        bridge_kwargs = consolidate_kwargs(
            kwargs_iterable=[
                get_function_default_arguments(adapter_class),
                model_setup_info.standard_bridge_kwargs,
                {"transforms": model_setup_info.transforms},
                kwargs,
            ],
            keywords=get_function_argument_names(
                function=adapter_class, omit=["experiment", "search_space", "data"]
            ),
        )

        # Create adapter with the consolidated kwargs.
        adapter = adapter_class(
            search_space=search_space or none_throws(experiment).search_space,
            experiment=experiment,
            data=data,
            generator=generator,
            **bridge_kwargs,
        )

        if model_setup_info.not_saved_model_kwargs:
            for key in model_setup_info.not_saved_model_kwargs:
                model_kwargs.pop(key, None)

        # Store all kwargs on adapter, to be saved on generator run.
        adapter._set_kwargs_to_save(
            model_key=self.value,
            model_kwargs=_encode_callables_as_references(model_kwargs),
            bridge_kwargs=_encode_callables_as_references(bridge_kwargs),
        )
        return adapter

    def view_defaults(self) -> tuple[dict[str, Any], dict[str, Any]]:
        """Obtains the default keyword arguments for the model and the adapter
        specified through the Generators enum, for ease of use in notebook environment,
        since models and adapters cannot be inspected directly through the enum.

        Returns:
            A tuple of default keyword arguments for the model and the adapter.
        """
        model_setup_info = none_throws(self.model_key_to_model_setup.get(self.value))
        return (
            self._get_model_kwargs(info=model_setup_info),
            self._get_bridge_kwargs(info=model_setup_info),
        )

    def view_kwargs(self) -> tuple[dict[str, Any], dict[str, Any]]:
        """Obtains annotated keyword arguments that the model and the adapter
        (corresponding to a given member of the Generators enum) constructors expect.

        Returns:
            A tuple of annotated keyword arguments for the model and the adapter.
        """
        model_class = self.model_class
        adapter_class = self.adapter_class
        return (
            {kw: p.annotation for kw, p in signature(model_class).parameters.items()},
            {kw: p.annotation for kw, p in signature(adapter_class).parameters.items()},
        )

    @staticmethod
    def _get_model_kwargs(
        info: ModelSetup, kwargs: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        return consolidate_kwargs(
            [get_function_default_arguments(info.model_class), kwargs],
            keywords=get_function_argument_names(info.model_class),
        )

    @staticmethod
    def _get_bridge_kwargs(
        info: ModelSetup, kwargs: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        return consolidate_kwargs(
            [
                get_function_default_arguments(info.adapter_class),
                info.standard_bridge_kwargs,
                {"transforms": info.transforms},
                kwargs,
            ],
            keywords=get_function_argument_names(
                info.adapter_class, omit=["experiment", "search_space", "data"]
            ),
        )


class Generators(GeneratorRegistryBase):
    """Registry of available models.

    Uses MODEL_KEY_TO_MODEL_SETUP to retrieve settings for model and adapter,
    by the key stored in the enum value.

    To instantiate a model in this enum, simply call an enum member like so:
    `Generators.SOBOL(search_space=search_space)` or
    `Generators.BOTORCH(experiment=experiment, data=data)`. Keyword arguments
    specified to the call will be passed into the model or the adapter
    constructors according to their keyword.

    For instance, `Generators.SOBOL(search_space=search_space, scramble=False)`
    will instantiate a `RandomAdapter(search_space=search_space)`
    with a `SobolGenerator(scramble=False)` underlying model.

    NOTE: If you deprecate a model, please add its replacement to
    `ax.storage.json_store.decoder._DEPRECATED_MODEL_TO_REPLACEMENT` to ensure
    backwards compatibility of the storage layer.
    """

    SOBOL = "Sobol"
    FACTORIAL = "Factorial"
    SAASBO = "SAASBO"
    SAAS_MTGP = "SAAS_MTGP"
    THOMPSON = "Thompson"
    LEGACY_BOTORCH = "Legacy_GPEI"
    BOTORCH_MODULAR = "BoTorch"
    EMPIRICAL_BAYES_THOMPSON = "EB"
    EB_ASHR = "EB_Ashr"
    UNIFORM = "Uniform"
    ST_MTGP = "ST_MTGP"
    BO_MIXED = "BO_MIXED"
    CONTEXT_SACBO = "Contextual_SACBO"


class ModelsMetaClass(type):
    """Metaclass to override `__getattr__` for the Models class."""

    def __getattr__(self, name: str) -> None:
        raise DeprecationWarning(
            "Models is deprecated, use `ax.adapter.registry.Generators` instead."
        )


class Models(metaclass=ModelsMetaClass):
    """This is deprecated. Use Generators instead."""

    pass


def _extract_model_state_after_gen(
    generator_run: GeneratorRun, model_class: type[Generator]
) -> dict[str, Any]:
    """Extracts serialized post-generation model state from a generator run and
    deserializes it.
    """
    serialized_model_state = generator_run._model_state_after_gen or {}
    return model_class.deserialize_state(serialized_model_state)


def _encode_callables_as_references(kwarg_dict: dict[str, Any]) -> dict[str, Any]:
    """Converts callables to references of form <module>.<qualname>, and returns
    the resulting dictionary.
    """
    return {
        k: (
            {"is_callable_as_path": True, "value": callable_to_reference(v)}
            if isfunction(v)
            else v
        )
        for k, v in kwarg_dict.items()
    }


def _decode_callables_from_references(kwarg_dict: dict[str, Any]) -> dict[str, Any]:
    """Retrieves callables from references of form <module>.<qualname>, and returns
    the resulting dictionary.
    """
    return {
        k: (
            callable_from_reference(assert_is_instance(v.get("value"), str))
            if isinstance(v, dict) and v.get("is_callable_as_path", False)
            else v
        )
        for k, v in kwarg_dict.items()
    }



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.benchmark.benchmark_method import BenchmarkMethod
from ax.benchmark.methods.sobol import get_sobol_generation_strategy
from ax.utils.common.testutils import TestCase
from pyre_extensions import none_throws


class TestBenchmarkMethod(TestCase):
    def test_benchmark_method(self) -> None:
        gs = get_sobol_generation_strategy()
        method = BenchmarkMethod(name="Sobol10", generation_strategy=gs)
        self.assertEqual(method.name, "Sobol10")

        # test that `fit_tracking_metrics` has been correctly set to False
        for step in method.generation_strategy._steps:
            self.assertFalse(none_throws(step.model_kwargs).get("fit_tracking_metrics"))
        self.assertEqual(method.timeout_hours, 4)

        method = BenchmarkMethod(generation_strategy=gs, timeout_hours=10)
        self.assertEqual(method.name, method.generation_strategy.name)
        self.assertEqual(method.timeout_hours, 10)

        # test that instantiation works with node-based strategies
        method = BenchmarkMethod(name="Sobol10", generation_strategy=gs)
        for node in method.generation_strategy._nodes:
            self.assertFalse(
                none_throws(node.generator_spec_to_gen_from.model_kwargs).get(
                    "fit_tracking_metrics"
                )
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import itertools
from typing import Mapping, Sequence

from ax.adapter.base import Adapter
from ax.adapter.torch import TorchAdapter
from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCardGroup
from ax.analysis.plotly.arm_effects import ArmEffectsPlot
from ax.analysis.plotly.bandit_rollout import BanditRollout
from ax.analysis.plotly.objective_p_feasible_frontier import (
    OBJ_PFEAS_CARDGROUP_SUBTITLE,
    ObjectivePFeasibleFrontierPlot,
)
from ax.analysis.plotly.scatter import (
    SCATTER_CARDGROUP_SUBTITLE,
    SCATTER_CARDGROUP_TITLE,
    ScatterPlot,
)
from ax.analysis.summary import Summary
from ax.analysis.utils import extract_relevant_adapter
from ax.core.arm import Arm
from ax.core.base_trial import TrialStatus
from ax.core.batch_trial import BatchTrial
from ax.core.experiment import Experiment
from ax.core.outcome_constraint import ScalarizedOutcomeConstraint
from ax.core.utils import is_bandit_experiment
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from pyre_extensions import override

RESULTS_CARDGROUP_TITLE = "Results Analysis"

RESULTS_CARDGROUP_SUBTITLE = (
    "Result Analyses provide a high-level overview of the results of the optimization "
    "process so far with respect to the metrics specified in experiment design."
)

ARM_EFFECTS_PAIR_CARDGROUP_TITLE = (
    "Metric Effects: Predicted and observed effects for all arms in the experiment"
)
ARM_EFFECTS_PAIR_CARDGROUP_SUBTITLE = (
    "These pair of plots visualize the metric effects for each arm, with the Ax "
    "model predictions on the left and the raw observed data on the right. The "
    "predicted effects apply shrinkage for noise and adjust for non-stationarity "
    "in the data, so they are more representative of the reproducible effects that "
    "will manifest in a long-term validation experiment. "
)


class ResultsAnalysis(Analysis):
    """
    An Analysis that provides a high-level overview of the results of the optimization
    process so far, e.g. effects on all arms. It produces an analysis card group.
    """

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardGroup:
        # Ensure we have an Experiment provided by the user to extract the relevant
        # metric names from.
        if experiment is None:
            raise UserInputError("ResultsAnalysis requires an Experiment.")

        # If the Experiment has an OptimizationConfig set, extract the objective and
        # constraint names.
        objective_names = []
        constraint_names = []
        if (optimization_config := experiment.optimization_config) is not None:
            objective_names = optimization_config.objective.metric_names
            for oc in optimization_config.outcome_constraints:
                if isinstance(oc, ScalarizedOutcomeConstraint):
                    constraint_names.extend([m.name for m in oc.metrics])
                else:
                    constraint_names.append(oc.metric.name)

        relevant_adapter = extract_relevant_adapter(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        # Relativize the effects if the status quo is set and there are BatchTrials
        # present.
        relativize = experiment.status_quo is not None and any(
            isinstance(trial, BatchTrial) for trial in experiment.trials.values()
        )
        # Compute both observed and modeled effects for each objective and constraint.
        arm_effect_pair_group = (
            ArmEffectsPair(
                metric_names=[*objective_names, *constraint_names],
                relativize=relativize,
            ).compute_or_error_card(
                experiment=experiment,
                generation_strategy=generation_strategy,
                adapter=adapter,
            )
            if len(objective_names) > 0
            else None
        )

        # If there are multiple objectives, compute scatter plots of each combination
        # of two objectives.
        objective_scatter_group = (
            AnalysisCardGroup(
                name="Objective Scatter Plots",
                title=SCATTER_CARDGROUP_TITLE + " (Objectives)",
                subtitle=SCATTER_CARDGROUP_SUBTITLE,
                children=[
                    ScatterPlot(
                        x_metric_name=x,
                        y_metric_name=y,
                        relativize=relativize,
                    ).compute_or_error_card(
                        experiment=experiment,
                        generation_strategy=generation_strategy,
                        adapter=adapter,
                    )
                    for x, y in itertools.combinations(objective_names, 2)
                ],
            )
            if len(objective_names) > 1
            else None
        )

        # If there are objectives and constraints, compute scatter plots of each
        # objective versus each constraint. Objectives are always plotted on the x-
        # axis and constraints on the y-axis.
        constraint_scatter_group = (
            AnalysisCardGroup(
                name="Constraint Scatter Plots",
                title=SCATTER_CARDGROUP_TITLE + " (Constraints)",
                subtitle=SCATTER_CARDGROUP_SUBTITLE,
                children=[
                    ScatterPlot(
                        x_metric_name=objective_name,
                        y_metric_name=constraint_name,
                        relativize=relativize,
                    ).compute_or_error_card(
                        experiment=experiment,
                        generation_strategy=generation_strategy,
                        adapter=adapter,
                    )
                    for objective_name in objective_names
                    for constraint_name in constraint_names
                ],
            )
            if len(objective_names) > 0 and len(constraint_names) > 0
            else None
        )

        objective_p_feasible_group = None
        if (
            len(objective_names) == 1
            and len(constraint_names) > 0
            # check that the adapter has a BoTorchGenerator
            and (
                relevant_adapter is not None
                and isinstance(relevant_adapter, TorchAdapter)
                and isinstance(relevant_adapter.generator, BoTorchGenerator)
            )
        ):
            objective_p_feasible_group = AnalysisCardGroup(
                name="Objective vs P(feasible)",
                title=(
                    "Model-Estimated Pareto-Frontier Between the Objective"
                    " and the Probability of Satisfying the Constraints"
                ),
                subtitle=OBJ_PFEAS_CARDGROUP_SUBTITLE,
                children=[
                    ObjectivePFeasibleFrontierPlot(
                        relativize=relativize
                    ).compute_or_error_card(
                        experiment=experiment,
                        generation_strategy=generation_strategy,
                        adapter=adapter,
                    )
                ],
            )

        # Produce a parallel coordinates plot for each objective.
        # TODO: mpolson mgarrard bring back parallel coordinates after fixing
        # objective_parallel_coordinates_group = (
        #     AnalysisCardGroup(
        #         name="Objective Parallel Coordinates Plots",
        #         children=[
        #             ParallelCoordinatesPlot(
        #                 metric_name=metric_name
        #             ).compute_or_error_card(
        #                 experiment=experiment,
        #                 generation_strategy=generation_strategy,
        #                 adapter=adapter,
        #             )
        #             for metric_name in objective_names
        #         ],
        #     )
        #     if len(objective_names) > 0
        #     else None
        # )

        # Add BanditRollout for experiments with specific generation strategy
        bandit_rollout_card = (
            BanditRollout().compute_or_error_card(
                experiment=experiment,
                generation_strategy=generation_strategy,
                adapter=adapter,
            )
            if generation_strategy
            and is_bandit_experiment(generation_strategy_name=generation_strategy.name)
            else None
        )

        summary = Summary().compute_or_error_card(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        return self._create_analysis_card_group(
            title=RESULTS_CARDGROUP_TITLE,
            subtitle=RESULTS_CARDGROUP_SUBTITLE,
            children=[
                child
                for child in (
                    arm_effect_pair_group,
                    objective_p_feasible_group,
                    objective_scatter_group,
                    constraint_scatter_group,
                    bandit_rollout_card,
                    summary,
                )
                if child is not None
            ],
        )


class ArmEffectsPair(Analysis):
    """
    Compute two ArmEffectsPlots in a single AnalysisCardGroup, one plotting model
    predictions and one plotting raw observed data.
    """

    def __init__(
        self,
        metric_names: Sequence[str] | None = None,
        relativize: bool = False,
        trial_index: int | None = None,
        trial_statuses: Sequence[TrialStatus] | None = None,
        additional_arms: Sequence[Arm] | None = None,
        labels: Mapping[str, str] | None = None,
        show_cumulative_best: bool = False,
    ) -> None:
        """
        Args:
            metric_names: The names of the metrics to include in the plot. If not
                specified, all metrics in the experiment will be used.
            relativize: Whether to relativize the effects of each arm against the status
                quo arm. If multiple status quo arms are present, relativize each arm
                against the status quo arm from the same trial.
            trial_index: If present, only use arms from the trial with the given index.
            additional_arms: If present, include these arms in the plot in addition to
                the arms in the experiment. These arms will be marked as belonging to a
                trial with index -1.
            labels: A mapping from metric names to labels to use in the plot. If a label
                is not provided for a metric, the metric name will be used.
            show_cumulative_best: Whether to draw a line through the best point seen so
                far during the optimization.
        """

        self.metric_names = metric_names
        self.relativize = relativize
        self.trial_index = trial_index
        self.trial_statuses = trial_statuses
        self.additional_arms = additional_arms
        self.labels: Mapping[str, str] = labels or {}
        self.show_cumulative_best = show_cumulative_best

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardGroup:
        if experiment is None:
            raise UserInputError("ArmEffectsPlot requires an Experiment.")

        relevant_adapter = extract_relevant_adapter(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        pairs: list[AnalysisCardGroup] = []
        for metric_name in self.metric_names or [*experiment.metrics.keys()]:
            # TODO: Test for no effects and render a message instead of a flat line.
            predicted_analysis = ArmEffectsPlot(
                metric_names=[metric_name],
                use_model_predictions=True,
                relativize=self.relativize,
                trial_index=self.trial_index,
                trial_statuses=self.trial_statuses,
                additional_arms=self.additional_arms,
                labels=self.labels,
                show_cumulative_best=self.show_cumulative_best,
            )

            raw_analysis = ArmEffectsPlot(
                metric_names=[metric_name],
                use_model_predictions=False,
                relativize=self.relativize,
                trial_index=self.trial_index,
                trial_statuses=self.trial_statuses,
                additional_arms=self.additional_arms,
                labels=self.labels,
                show_cumulative_best=self.show_cumulative_best,
            )

            pair = AnalysisCardGroup(
                name=f"ArmEffects Pair {metric_name}",
                title=f"Metric Effects Pair for {metric_name}",
                subtitle=None,
                children=[
                    predicted_analysis.compute_or_error_card(
                        experiment=experiment,
                        generation_strategy=generation_strategy,
                        adapter=relevant_adapter,
                    ),
                    raw_analysis.compute_or_error_card(experiment=experiment),
                ],
            )

            pairs.append(pair)

        return self._create_analysis_card_group(
            title=ARM_EFFECTS_PAIR_CARDGROUP_TITLE,
            subtitle=ARM_EFFECTS_PAIR_CARDGROUP_SUBTITLE,
            children=pairs,
        )
